import calendar
from datetime import date
import traceback
from typing import List, Tuple

from dask.diagnostics import ProgressBar
import numpy as np
import pandas as pd
import xarray as xr

from .coast import Coast
from .logging_util import info, warn, error


class Climatology(Coast):  # TODO All abstract methods should be implemented
    """
    A Python class containing methods for lazily creating climatologies of
    NEMO data (or any xarray datasets) and writing to file. Also for resampling
    methods.
    """

    def __init__(self):
        return  # TODO Super __init__ should be called at some point

    @staticmethod
    def make_climatology(
        ds, output_frequency, monthly_weights=False, time_var_name="time", time_dim_name="t_dim", fn_out=None
    ):
        """
        Calculates a climatology for all variables in a supplied dataset.
        The resulting xarray dataset will NOT be loaded to RAM. Instead,
        it is a set of dask operations. To load to RAM use, e.g. .compute().
        However, if the original data was large, this may take a long time and
        a lot of memory. Make sure you have the available RAM or chunking
        and parallel processes are specified correctly.

        Otherwise, it is recommended that you access the climatology data
        in an indexed way. I.E. compute only at specific parts of the data
        are once.

        The resulting cliamtology dataset can be written to disk using
        .to_netcdf(). Again, this may take a while for larger datasets.

        ds :: xarray dataset object from a Coast object.
        output_frequency :: any xarray groupby string. i.e:
            'month'
            'season'
        time_var_name :: the string name of the time variable in dataset
        time_dim_name :: the string name of the time dimension variable in dataset
        fn_out :: string defining full output netcdf file path and name.
        """

        frequency_str = time_var_name + "." + output_frequency
        info("Calculating climatological mean")

        if monthly_weights:
            month_length = ds[time_var_name].dt.days_in_month
            grouped = month_length.groupby(frequency_str)
            weights = grouped / grouped.sum()
            ds_mean = (ds * weights).groupby(frequency_str).sum(dim=time_dim_name)
        else:
            ds_mean = xr.Dataset()
            for var_name, da in ds.data_vars.items():
                try:
                    da_mean = da.groupby(frequency_str).mean(dim=time_dim_name, skipna=True)
                    ds_mean[var_name] = da_mean
                except ArithmeticError:
                    error(f"Skipped mean calculation for {var_name} due to error: {traceback.format_exc()}")

        if fn_out is not None:
            info("Saving to file. May take some time..")
            with ProgressBar():
                ds_mean.to_netcdf(fn_out)

        return ds_mean

    @staticmethod
    def _get_date_ranges(years: List[int], month_periods: List[Tuple[int, int]]) -> List[Tuple[date, date]]:
        """Calculates a list of datetime date ranges for a given list of years and a specified start/end month.

        Args:
            years (list): A list of years to calculate date ranges for.
            month_periods (list): A list containing tuples of start and end month integers.
            (i.e. [(3,5),(12, 2)] is Mar -> May, Dec -> Feb). Must be in chronological order.

            Returns:
                date_ranges (list): A list of tuples, each containing a start and end datetime.date object.
        """
        date_ranges = []
        for y in sorted(set(years)):
            y = int(y)
            for period in month_periods:
                start = period[0]
                end = period[1]
                begin_date = date(y, start, 1)
                if start > end:
                    end_day = calendar.monthrange(y + 1, end)[1]
                    end_date = date(y + 1, end, end_day)
                else:
                    end_day = calendar.monthrange(y, end)[1]
                    end_date = date(y, end, end_day)
                date_ranges.append((begin_date, end_date))
        return date_ranges

    @classmethod
    def multiyear_averages(
        cls, ds: xr.Dataset, month_periods: List[Tuple[int, int]], time_var: str = "time", time_dim: str = "t_dim"
    ) -> xr.Dataset:
        """Calculate multiyear means for all Data variables in a dataset between a given start and end month.

        Args:
            ds (xr.Dataset): xarray dataset containing data.
            month_periods (list): A list containing tuples of start and end month integers.
            (i.e. [(3,5),(12, 2)] is Mar -> May, Dec -> Feb). Must be in chronological order.
            The seasons module can be used for convenience (e.g. seasons.WINTER, seasons.ALL etc. )
            time_var (str): String representing the time variable name within the dataset.
            time_dim (str): String representing the time dimension name within the dataset.
        returns:
            ds_mean (xr.Dataset): A new dataset containing mean averages for each data variable across all years and
            date periods. Indexed by the multi-index 'year_period' (i.e. (2000, 'Dec-Feb')).
        """

        time_dim_da = ds[f"{time_dim}"]
        time_var_da = ds[f"{time_var}"]

        new_ds = ds
        # If time dimension isn't np.datetime64 but time variable is, then swap time variable to be the dimension.
        # There should be a 1 to 1 mapping between time dimension values and time variable values.
        # A datetime type is required for slicing dates over a dimension using xarray's sel() method.
        if not np.issubdtype(time_dim_da.dtype, np.datetime64) and np.issubdtype(time_var_da.dtype, np.datetime64):
            warn("Time dimension is not np.datatime64 but time variable is. Swapping time dimension for data variable.")
            # Swap time_var with time_dim.
            new_ds = ds.swap_dims({f"{time_dim}": f"{time_var}"})
            time_dim = time_var
        elif not np.issubdtype(time_dim_da.dtype, np.datetime64) and not np.issubdtype(
            time_var_da.dtype, np.datetime64
        ):
            # Slicing will most likely fail for non np.datetime64 datatypes.
            warn("Neither time dimension or time variable data are np.datetime64. Time slicing may fail.")

        # Get years of data.
        data_years = list(new_ds[f"{time_dim}.year"].data)
        # Append first year - 1, to account for possible WINTER month data of year prior to data beginning.
        data_years.insert(0, data_years[0] - 1)

        # Generate date ranges from years and given month periods.
        date_ranges = Climatology._get_date_ranges(data_years, month_periods)

        # Extract data from dataset between these date ranges and index each range with a common multi-index.
        datasets = []
        year_index = []
        month_index = []
        for date_range in date_ranges:
            sel_args = {f"{time_dim}": slice(date_range[0], date_range[1])}
            filtered = new_ds.sel(**sel_args)
            datasets.append(filtered)
            year_index = year_index + ([date_range[0].year] * filtered.sizes[time_dim])
            month_label = f"{calendar.month_abbr[date_range[0].month]}-{calendar.month_abbr[date_range[1].month]}"
            month_index = month_index + ([month_label] * filtered.sizes[time_dim])

        # New dataset built from extracted data between date ranges.
        filtered = xr.concat(datasets, dim=time_dim)
        # Data from same date range use common year-period multi-index so they can be grouped together.
        period_idx = pd.MultiIndex.from_arrays([year_index, month_index], names=("year", "period"))
        filtered.coords["year_period"] = (f"{time_dim}", period_idx)

        # For each data variable, group on year-period multi-index and find the mean.
        # New dataset containing means across date ranges is returned.
        ds_mean = xr.Dataset()
        for var_name, da in filtered.data_vars.items():
            try:
                # Apply .mean() to grouped data.
                # skipna flag used to ignore NaN values.
                da_mean = da.groupby("year_period").mean(dim=time_dim, skipna=True)
                ds_mean[f"{var_name}"] = da_mean
            except ArithmeticError:
                warn(f"Skipped mean calculation for {var_name} due to error: {traceback.format_exc()}")
        return ds_mean



"""Profile Class"""
from .index import Indexed
import numpy as np
import xarray as xr
from . import general_utils, plot_util
import matplotlib.pyplot as plt
import glob
import datetime
from .logging_util import get_slug, debug, info, warn, warning
from typing import Union
from pathlib import Path
import pandas as pd


class Profile(Indexed):
    """
    INDEXED type class for storing data from a CTD Profile (or similar
    down and up observations). The structure of the class is based around having
    discrete profile locations with independent depth dimensions and coords.
    The class dataset should contain two dimensions:

        > id_dim      :: The profiles dimension. Each element of this dimension
                     contains data (e.g. cast) for an individual location.
        > z_dim   :: The dimension for depth levels. A profile object does not
                     need to have shared depths, so NaNs might be used to
                     pad any depth array.

    Alongside these dimensions, the following minimal coordinates should also
    be available:

        > longitude (id_dim)   :: 1D array of longitudes, one for each id_dim
        > latitude  (id_dim)   :: 1D array of latitudes, one for each id_dim
        > time      (id_dim)   :: 1D array of times, one for each id_dim
        > depth     (id_dim, z_dim)  :: 2D array of depths, with different depth
                                    levels being provided for each profile.
                                    Note that these depth levels need to be
                                    stored in a 2D array, so NaNs can be used
                                    to pad out profiles with shallower depths.
        > id_name   (id_dim)   :: [Optional] Name of id_dim/case or id_dim number.

    You may create an empty profile object by using profile = coast.Profile().
    You may then add your own dataset to the object profile or use one of the
    functions within Profile() for reading common profile datasets:

        > read_en4()
        > read_wod()

    Optionally, you may pass a dataset to the Profile object on creation:

        profile = coast.Profile(dataset = profile_dataset)

    A config file can also be provided, in which case any netcdf read functions
    will rename dimensions and variables as dictated.
    """

    def __init__(self, dataset=None, config: Union[Path, str] = None):
        """Initialization and file reading. You may initialize

        Args:
            config (Union[Path, str]): path to json config file.
        """
        debug(f"Creating a new {get_slug(self)}")
        self.config = config
        super().__init__(self.config)

        # If dataset is provided, put inside this object
        if dataset is not None:
            self.dataset = dataset
            self.apply_config_mappings()

        debug(f"{get_slug(self)} initialised")

    def read_en4(self, fn_en4, chunks: dict = {}, multiple=False) -> None:
        """
        Reads a single or multiple EN4 netCDF files into the COAsT profile
        data structure.

        Parameters
        ----------
        fn_en4 : TYPE
            path to data file.
        chunks : dict, optional
            Chunking specification
        multiple : TYPE, optional
            True if reading multiple files otherwise False

        Returns
        -------
        None. Populates dataset within Profile object.
        """

        # If not multiple then just read the netcdf file
        if not multiple:
            self.dataset = xr.open_dataset(fn_en4, chunks=chunks)

        # If multiple, then we have to get all file names and read them in a
        # loop, followed by concatenation
        else:
            # Check a list is provided
            if type(fn_en4) is not list:
                fn_en4 = [fn_en4]

            # Use glob to get a list of file paths from input
            file_to_read = []
            for file in fn_en4:
                if "*" in file:
                    wildcard_list = glob.glob(file)
                    file_to_read = file_to_read + wildcard_list
                else:
                    file_to_read.append(file)

            # Reorder files to read
            file_to_read = np.array(file_to_read)
            dates = [ff[-9:-3] for ff in file_to_read]
            dates = [datetime.datetime(int(dd[0:4]), int(dd[4:6]), 1) for dd in dates]
            sort_ind = np.argsort(dates)
            file_to_read = file_to_read[sort_ind]

            # Loop over all files, open them and concatenation them into one
            for ff in range(0, len(file_to_read)):
                file = file_to_read[ff]
                data_tmp = xr.open_dataset(file, chunks=chunks)
                if ff == 0:
                    self.dataset = data_tmp
                else:
                    self.dataset = xr.concat((self.dataset, data_tmp), dim="N_PROF")

        # Apply config settings
        self.apply_config_mappings()

    def read_wod(self, fn_wod, chunks: dict = {}) -> None:
        """Reads a single World Ocean Database netCDF files into the COAsT profile data structure.

        Args:
            fn_wod (str): path to data file
            chunks (dict): chunks
        """
        self.dataset = xr.open_dataset(fn_wod, chunks=chunks)
        self.apply_config_mappings()

    """======================= Manipulate ======================="""

    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Get a subset of this Profile() object in a spatial box.

        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]

        return: A new profile object containing subsetted data
        """
        ind = general_utils.subset_indices_lonlat_box(
            self.dataset.longitude, self.dataset.latitude, lonbounds[0], lonbounds[1], latbounds[0], latbounds[1]
        )
        return Profile(dataset=self.dataset.isel(id_dim=ind[0]))

    """======================= Plotting ======================="""

    def plot_profile(self, var: str, profile_indices=None):

        fig = plt.figure(figsize=(7, 10))

        if profile_indices is None:
            profile_indices = np.arange(0, self.dataset.dims["id_dim"])
            pass

        for ii in profile_indices:
            prof_var = self.dataset[var].isel(id_dim=ii)
            prof_depth = self.dataset.depth.isel(id_dim=ii)
            ax = plt.plot(prof_var, prof_depth)

        plt.gca().invert_yaxis()
        plt.xlabel(var + "(" + self.dataset[var].units + ")")
        plt.ylabel("Depth (" + self.dataset.depth.units + ")")
        plt.grid()
        return fig, ax

    def plot_map(self, var_str=None):

        profiles = self.dataset

        if var_str is None:
            fig, ax = plot_util.geo_scatter(profiles.longitude.values, profiles.latitude.values, s=5)
        else:
            c = profiles[var_str]
            fig, ax = plot_util.geo_scatter(profiles.longitude.values, profiles.latitude.values, c=c, s=5)
        return fig, ax

    def plot_ts_diagram(self, profile_index, var_t="potential_temperature", var_s="practical_salinity"):

        profile = self.dataset.isel(id_dim=profile_index)
        temperature = profile[var_t].values
        salinity = profile[var_s].values
        depth = profile.depth.values
        fig, ax = plot_util.ts_diagram(temperature, salinity, depth)

        return fig, ax

    """======================= Model Comparison ======================="""

    def process_en4(self, sort_time=True):
        """
        VERSION 1.4 (05/07/2021)

        PREPROCESSES EN4 data ready for comparison with model data.
        This routine will cut out a desired geographical box of EN4 data and
        then apply quality control according to the available flags in the
        netCDF files. Quality control happens in two steps:
            1. Where a whole data profile is flagged, it is completely removed
               from the dataset
            2. Where a single datapoint is rejected in either temperature or
               salinity, it is set to NaN.
        This routine attempts to use xarray/dask chunking magic to keep
        memory useage low however some memory is still needed for loading
        flags etc. May be slow if using large EN4 datasets.

        Routine will return a processed profile object dataset and can write
        the new dataset to file if fn_out is defined. If saving to the
        PROFILE object, be aware that DASK computations will not have happened
        and will need to be done using .load(), .compute() or similar before
        accessing the values. IF using multiple EN4 files or large dataset,
        make sure you have chunked the data over N_PROF dimension.

        INPUTS
         fn_out (str)      : Full path to a desired output file. If unspecified
                             then nothing is written.

        EXAMPLE USEAGE:
         profile = coast.PROFILE()
         profile.read_EN4(fn_en4, chunks={'N_PROF':10000})
         fn_out = '~/output_file.nc'
         new_profile = profile.preprocess_en4(fn_out = fn_out,
                                              lonbounds = [-10, 10],
                                              latbounds = [45, 65])
        """

        ds = self.dataset

        # Load in the quality control flags
        debug(f" Applying QUALITY CONTROL to EN4 data...")
        ds.qc_flags_profiles.load()

        # This line reads converts the QC integer to a binary string.
        # Each bit of this string is a different QC flag. Which flag is which can
        # be found on the EN4 website:
        # https://www.metoffice.gov.uk/hadobs/en4/en4-0-2-profile-file-format.html
        qc_str = [np.binary_repr(ds.qc_flags_profiles.values[pp]).zfill(30)[::-1] for pp in range(ds.sizes["id_dim"])]

        # Determine indices of the profiles that we want to keep
        reject_tem_prof = np.array([int(qq[0]) for qq in qc_str], dtype=bool)
        reject_sal_prof = np.array([int(qq[1]) for qq in qc_str], dtype=bool)
        reject_both_prof = np.logical_and(reject_tem_prof, reject_sal_prof)
        ds["reject_tem_prof"] = (["id_dim"], reject_tem_prof)
        ds["reject_sal_prof"] = (["id_dim"], reject_sal_prof)
        debug("     >>> QC: Completely rejecting {0} / {1} id_dims".format(np.sum(reject_both_prof), ds.dims["id_dim"]))

        # Subset profile dataset to remove profiles that are COMPLETELY empty
        ds = ds.isel(id_dim=~reject_both_prof)
        reject_tem_prof = reject_tem_prof[~reject_both_prof]
        reject_sal_prof = reject_sal_prof[~reject_both_prof]

        # Get new QC flags array
        qc_lev = ds.qc_flags_levels.values

        debug(f" QC: Additional profiles converted to NaNs: ")
        debug(f"     >>> {0} temperature profiles ".format(np.sum(reject_tem_prof)))
        debug(f"     >>> {0} salinity profiles ".format(np.sum(reject_sal_prof)))

        #
        reject_tem_lev = np.zeros((ds.dims["id_dim"], ds.dims["z_dim"]), dtype=bool)
        reject_sal_lev = np.zeros((ds.dims["id_dim"], ds.dims["z_dim"]), dtype=bool)

        int_tem, int_sal, int_both = self.calculate_all_en4_qc_flags()
        for ii in range(len(int_tem)):
            reject_tem_lev[qc_lev == int_tem[ii]] = 1
        for ii in range(len(int_sal)):
            reject_sal_lev[qc_lev == int_sal[ii]] = 1
        for ii in range(len(int_both)):
            reject_tem_lev[qc_lev == int_both[ii]] = 1
            reject_sal_lev[qc_lev == int_both[ii]] = 1

        ds["reject_tem_datapoint"] = (["id_dim", "z_dim"], reject_tem_lev)
        ds["reject_sal_datapoint"] = (["id_dim", "z_dim"], reject_sal_lev)

        debug(f"MASKING rejected datapoints, replacing with NaNs...")
        ds["temperature"] = xr.where(~reject_tem_lev, ds["temperature"], np.nan)
        ds["potential_temperature"] = xr.where(~reject_tem_lev, ds["temperature"], np.nan)
        ds["practical_salinity"] = xr.where(~reject_tem_lev, ds["practical_salinity"], np.nan)

        if sort_time:
            debug(f"Sorting Time Dimension...")
            ds = ds.sortby("time")

        debug(f"Finished processing data. Returning new Profile object.")

        return_prof = Profile()
        return_prof.dataset = ds
        return return_prof

    @classmethod
    def calculate_all_en4_qc_flags(cls):
        """
        Brute force method for identifying all rejected points according to
        EN4 binary integers. It can be slow to convert large numbers of integers
        to a sequence of bits and is actually quicker to just generate every
        combination of possible QC integers. That's what this routine does.
        Used in PROFILE.preprocess_en4().

        INPUTS
         NO INPUTS

        OUTPUTS
         qc_integers_tem  : Array of integers signifying the rejection of ONLY
                            temperature datapoints
         qc_integers_sal  : Array of integers signifying the rejection of ONLY
                            salinity datapoints
         qc_integers_both : Array of integers signifying the rejection of BOTH
                            temperature and salinity datapoints.
        """

        reject_tem_ind = 0
        reject_sal_ind = 1
        reject_tem_reasons = [2, 3, 8, 9, 10, 11, 12, 13, 14, 15, 16]
        reject_sal_reasons = [2, 3, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29]

        qc_integers_tem = []
        qc_integers_sal = []
        qc_integers_both = []
        n_tem_reasons = len(reject_tem_reasons)
        n_sal_reasons = len(reject_sal_reasons)
        bin_len = 30

        # IF reject_tem = 1, reject_sal = 0
        for ii in range(n_tem_reasons):
            bin_tmp = np.zeros(bin_len, dtype=int)
            bin_tmp[reject_tem_ind] = 1
            bin_tmp[reject_tem_reasons[ii]] = 1
            qc_integers_tem.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        # IF reject_tem = 0, reject_sal = 1
        for ii in range(n_sal_reasons):
            bin_tmp = np.zeros(bin_len, dtype=int)
            bin_tmp[reject_sal_ind] = 1
            bin_tmp[reject_sal_reasons[ii]] = 1
            qc_integers_sal.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        # IF reject_tem = 1, reject_sal = 1
        for tt in range(n_tem_reasons):
            for ss in range(n_sal_reasons):
                bin_tmp = np.zeros(bin_len, dtype=int)
                bin_tmp[reject_tem_ind] = 1
                bin_tmp[reject_sal_ind] = 1
                bin_tmp[reject_tem_reasons[tt]] = 1
                bin_tmp[reject_sal_reasons[ss]] = 1
                qc_integers_both.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        qc_integers_tem = list(set(qc_integers_tem))
        qc_integers_sal = list(set(qc_integers_sal))
        qc_integers_both = list(set(qc_integers_both))

        return qc_integers_tem, qc_integers_sal, qc_integers_both

    def obs_operator(self, gridded, mask_bottom_level=True):
        """
        VERSION 2.0 (04/10/2021)
        Author: David Byrne

        Does a spatial and time interpolation of a gridded object's data.
        A nearest neighbour approach is used for both interpolations. Both
        datasets (the Profile and Gridded objects) must contain longitude,
        latitude and time coordinates. This routine expects there to be a
        landmask variable in the gridded object. This is is not available,
        then place an array of zeros into the dataset, with dimensions
        (y_dim, x_dim).

        This routine will do the interpolation based on the chunking applied
        to the Gridded object. Please ensure you have the available memory to
        have an entire Gridded chunk loaded to memory. If multiple files are
        used, then using one chunk per file will be most efficient. Time
        chunking is generally the better option for this routine.

        INPUTS:
         gridded (Gridded)        : gridded object created on t-grid
         mask_bottom_level (bool) : Whether or not to mask any data below the
                                    model's bottom level. If True, then ensure
                                    the Gridded object's dataset contain's a
                                    bottom_level variable with dims
                                    (y_dim, x_dim).

        OUTPUTS:
         Returns a new PROFILE object containing a computed dataset of extracted
         profiles.
        """

        # Read EN4, then extract desired variables
        en4 = self.dataset
        gridded = gridded.dataset

        # CHECKS
        # 1. Check that bottom_level is in dataset if mask_bottom_level is True
        if mask_bottom_level:
            if "bottom_level" not in gridded.variables:
                raise ValueError(
                    "bottom_level not found in input dataset. Please ensure variable is present or set mask_bottom_level to False"
                )

        # Use only observations that are within model time window.
        en4_time = en4.time.values
        mod_time = gridded.time.values

        # SPATIAL indices - nearest neighbour
        ind_x, ind_y = general_utils.nearest_indices_2d(
            gridded["longitude"], gridded["latitude"], en4["longitude"], en4["latitude"], mask=gridded.landmask
        )
        debug(f"Spatial Indices Calculated")

        # TIME indices - model nearest to obs time
        en4_time = en4.time.values
        ind_t = [np.argmin(np.abs(mod_time - en4_time[tt])) for tt in range(en4.dims["id_dim"])]
        ind_t = xr.DataArray(ind_t)
        debug(f"Time Indices Calculated")

        # Find out which variables have both depth and profile
        # This is for applying the bottom_level mask later
        var_list = list(gridded.keys())
        bl_var_list = []
        for vv in var_list:
            cond1 = "z_dim" not in gridded[vv].dims
            if cond1:
                bl_var_list.append(vv)

        # Get chunks along the time dimension and determine whether chunks
        # are described by a single equal size, or a tuples of sizes
        time_chunks = gridded.chunks["t_dim"]
        time_dim = gridded.dims["t_dim"]
        start_ii = 0  # Starting index for loading data. Increments each loop
        count_ii = 0  # Counting index for allocating data. Increments 1 each loop

        while start_ii < time_dim:
            end_ii = start_ii + time_chunks[count_ii]
            debug(f"{0}: {1} > {2}".format(count_ii, start_ii, end_ii))

            # Determine which time indices lie in this chunk
            ind_in_chunk = np.logical_and(ind_t >= start_ii, ind_t < end_ii)

            # Check There are some indices at all
            if np.sum(ind_in_chunk) == 0:
                start_ii = end_ii
                if count_ii == 0:
                    mod_profiles = xr.Dataset()

                count_ii = count_ii + 1
                continue

            # Pull out x,y and t indices
            ind_x_in_chunk = ind_x[ind_in_chunk]
            ind_y_in_chunk = ind_y[ind_in_chunk]
            ind_t_in_chunk = ind_t[ind_in_chunk] - start_ii

            # Index a temporary chunk and read it to memory
            ds_tmp = gridded.isel(t_dim=np.arange(start_ii, end_ii)).load()

            # Index loaded chunk and rename dim_0 to profile
            ds_tmp_indexed = ds_tmp.isel(x_dim=ind_x_in_chunk, y_dim=ind_y_in_chunk, t_dim=ind_t_in_chunk)
            ds_tmp_indexed = ds_tmp_indexed.rename({"dim_0": "id_dim"})

            # Mask out all levels deeper than bottom_level
            # Here I have used set_coords() and reset_coords() to omit variables
            # with no z_dim from the masking. Otherwise xr.where expands these
            # dimensions into full 2D arrays.
            if mask_bottom_level:
                n_z_tmp = ds_tmp_indexed.dims["z_dim"]
                bl_array = ds_tmp_indexed.bottom_level.values
                z_index, bl_index = np.meshgrid(np.arange(0, n_z_tmp), bl_array)
                mask2 = xr.DataArray(z_index < bl_index, dims=["id_dim", "z_dim"])
                ds_tmp_indexed = ds_tmp_indexed.set_coords(bl_var_list)
                ds_tmp_indexed = ds_tmp_indexed.where(mask2)
                ds_tmp_indexed = ds_tmp_indexed.reset_coords(bl_var_list)

            # If not first iteration, concatenate this indexed chunk onto
            # final output dataset
            if count_ii == 0:
                mod_profiles = ds_tmp_indexed
            else:
                if len(mod_profiles.data_vars) == 0:
                    mod_profiles = xr.merge((mod_profiles, ds_tmp_indexed))
                else:
                    mod_profiles = xr.concat((mod_profiles, ds_tmp_indexed), dim="id_dim")

            # Update counters
            start_ii = end_ii
            count_ii = count_ii + 1

        # Put obs time into the output array
        mod_profiles["obs_time"] = (["id_dim"], en4_time)

        # Calculate interpolation distances
        interp_dist = general_utils.calculate_haversine_distance(
            en4.longitude, en4.latitude, mod_profiles.longitude, mod_profiles.latitude
        )
        mod_profiles["interp_dist"] = (["id_dim"], interp_dist.values)

        # Calculate interpolation time lags
        interp_lag = (mod_profiles.time.values - en4_time).astype("timedelta64[h]")
        mod_profiles["interp_lag"] = (["id_dim"], interp_lag)

        # Put x and y indices into dataset
        mod_profiles["nearest_index_x"] = (["id_dim"], ind_x.values)
        mod_profiles["nearest_index_y"] = (["id_dim"], ind_y.values)
        mod_profiles["nearest_index_t"] = (["id_dim"], ind_t.values)
        return Profile(dataset=mod_profiles)

    def process_en4(self, sort_time=True):
        """
        VERSION 1.4 (05/07/2021)

        PREPROCESSES EN4 data ready for comparison with model data.
        This routine will cut out a desired geographical box of EN4 data and
        then apply quality control according to the available flags in the
        netCDF files. Quality control happens in two steps:
            1. Where a whole data profile is flagged, it is completely removed
               from the dataset
            2. Where a single datapoint is rejected in either temperature or
               salinity, it is set to NaN.
        This routine attempts to use xarray/dask chunking magic to keep
        memory useage low however some memory is still needed for loading
        flags etc. May be slow if using large EN4 datasets.

        Routine will return a processed profile object dataset and can write
        the new dataset to file if fn_out is defined. If saving to the
        PROFILE object, be aware that DASK computations will not have happened
        and will need to be done using .load(), .compute() or similar before
        accessing the values. IF using multiple EN4 files or large dataset,
        make sure you have chunked the data over N_PROF dimension.

        INPUTS
         fn_out (str)      : Full path to a desired output file. If unspecified
                             then nothing is written.

        EXAMPLE USEAGE:
         profile = coast.PROFILE()
         profile.read_EN4(fn_en4, chunks={'N_PROF':10000})
         fn_out = '~/output_file.nc'
         new_profile = profile.preprocess_en4(fn_out = fn_out,
                                              lonbounds = [-10, 10],
                                              latbounds = [45, 65])
        """

        ds = self.dataset

        # REJECT profiles that are QC flagged.
        debug(f" Applying QUALITY CONTROL to EN4 data...")
        ds.qc_flags_profiles.load()

        # This line reads converts the QC integer to a binary string.
        # Each bit of this string is a different QC flag. Which flag is which can
        # be found on the EN4 website:
        # https://www.metoffice.gov.uk/hadobs/en4/en4-0-2-profile-file-format.html
        qc_str = [np.binary_repr(ds.qc_flags_profiles.values[pp]).zfill(30)[::-1] for pp in range(ds.dims["id_dim"])]

        # Determine indices of kept profiles
        reject_tem_prof = np.array([int(qq[0]) for qq in qc_str], dtype=bool)
        reject_sal_prof = np.array([int(qq[1]) for qq in qc_str], dtype=bool)
        reject_both_prof = np.logical_and(reject_tem_prof, reject_sal_prof)
        ds["reject_tem_prof"] = (["id_dim"], reject_tem_prof)
        ds["reject_sal_prof"] = (["id_dim"], reject_sal_prof)
        debug(
            "     >>> QC: Completely rejecting {0} / {1} profiles".format(np.sum(reject_both_prof), ds.dims["id_dim"])
        )

        ds = ds.isel(id_dim=~reject_both_prof)
        reject_tem_prof = reject_tem_prof[~reject_both_prof]
        reject_sal_prof = reject_sal_prof[~reject_both_prof]
        qc_lev = ds.qc_flags_levels.values

        debug(f" QC: Additional profiles converted to NaNs: ")
        debug(f"     >>> {0} temperature profiles ".format(np.sum(reject_tem_prof)))
        debug(f"     >>> {0} salinity profiles ".format(np.sum(reject_sal_prof)))

        reject_tem_lev = np.zeros((ds.dims["id_dim"], ds.dims["z_dim"]), dtype=bool)
        reject_sal_lev = np.zeros((ds.dims["id_dim"], ds.dims["z_dim"]), dtype=bool)

        int_tem, int_sal, int_both = self.calculate_all_en4_qc_flags()
        for ii in range(len(int_tem)):
            reject_tem_lev[qc_lev == int_tem[ii]] = 1
        for ii in range(len(int_sal)):
            reject_sal_lev[qc_lev == int_sal[ii]] = 1
        for ii in range(len(int_both)):
            reject_tem_lev[qc_lev == int_both[ii]] = 1
            reject_sal_lev[qc_lev == int_both[ii]] = 1

        ds["reject_tem_datapoint"] = (["id_dim", "z_dim"], reject_tem_lev)
        ds["reject_sal_datapoint"] = (["id_dim", "z_dim"], reject_sal_lev)

        debug(f"MASKING rejected datapoints, replacing with NaNs...")
        ds["temperature"] = xr.where(~reject_tem_lev, ds["temperature"], np.nan)
        ds["potential_temperature"] = xr.where(~reject_tem_lev, ds["temperature"], np.nan)
        ds["practical_salinity"] = xr.where(~reject_tem_lev, ds["practical_salinity"], np.nan)

        if sort_time:
            debug(f"Sorting Time Dimension...")
            ds = ds.sortby("time")

        debug(f"Finished processing data. Returning new Profile object.")

        return_prof = Profile()
        return_prof.dataset = ds
        return return_prof

    def calculate_all_en4_qc_flags(self):
        """
        Brute force method for identifying all rejected points according to
        EN4 binary integers. It can be slow to convert large numbers of integers
        to a sequence of bits and is actually quicker to just generate every
        combination of possible QC integers. That's what this routine does.
        Used in PROFILE.preprocess_en4().

        INPUTS
         NO INPUTS

        OUTPUTS
         qc_integers_tem  : Array of integers signifying the rejection of ONLY
                            temperature datapoints
         qc_integers_sal  : Array of integers signifying the rejection of ONLY
                            salinity datapoints
         qc_integers_both : Array of integers signifying the rejection of BOTH
                            temperature and salinity datapoints.
        """

        reject_tem_ind = 0
        reject_sal_ind = 1
        reject_tem_reasons = [2, 3, 8, 9, 10, 11, 12, 13, 14, 15, 16]
        reject_sal_reasons = [2, 3, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29]

        qc_integers_tem = []
        qc_integers_sal = []
        qc_integers_both = []
        n_tem_reasons = len(reject_tem_reasons)
        n_sal_reasons = len(reject_sal_reasons)
        bin_len = 30

        # IF reject_tem = 1, reject_sal = 0
        for ii in range(n_tem_reasons):
            bin_tmp = np.zeros(bin_len, dtype=int)
            bin_tmp[reject_tem_ind] = 1
            bin_tmp[reject_tem_reasons[ii]] = 1
            qc_integers_tem.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        # IF reject_tem = 0, reject_sal = 1
        for ii in range(n_sal_reasons):
            bin_tmp = np.zeros(bin_len, dtype=int)
            bin_tmp[reject_sal_ind] = 1
            bin_tmp[reject_sal_reasons[ii]] = 1
            qc_integers_sal.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        # IF reject_tem = 1, reject_sal = 1
        for tt in range(n_tem_reasons):
            for ss in range(n_sal_reasons):
                bin_tmp = np.zeros(bin_len, dtype=int)
                bin_tmp[reject_tem_ind] = 1
                bin_tmp[reject_sal_ind] = 1
                bin_tmp[reject_tem_reasons[tt]] = 1
                bin_tmp[reject_sal_reasons[ss]] = 1
                qc_integers_both.append(int("".join(str(jj) for jj in bin_tmp)[::-1], 2))

        qc_integers_tem = list(set(qc_integers_tem))
        qc_integers_sal = list(set(qc_integers_sal))
        qc_integers_both = list(set(qc_integers_both))

        return qc_integers_tem, qc_integers_sal, qc_integers_both

    """================Reshape to 2D================"""

    def reshape_2d(self, var_user_want):
        """
        OBSERVATION type class for reshaping World Ocean Data (WOD) or similar that
        contains 1D profiles (profile * depth levels)  into a 2D array.
        Note that its variable has its own dimention and in some profiles
        only some variables are present. WOD can be observed depth or a
        standard depth as regrided by NOAA.
           Args:
            > X     --      The variable (e.g,Temperatute, Salinity, Oxygen, DIC ..)
            > X_N    --     Dimensions of observed variable as 1D
                            (essentially number of obs variable = casts * osberved depths)
            > casts --      Dimension for locations of observations (ie. profiles)
            > z_N   --      Dimension for depth levels of all observations as 1D
                            (essentially number of depths = casts * osberved depths)
            > X_row_size -- Gives the vertical index (number of depths)
                            for each variable
        """

        """reshape the 1D variable into 2D variable (id_dimß, z_dim)    
        Args:
            profile       ::   The profile dimension. Called cast in WOD, 
                               common in all variables, but nans if 
                                a variable is not observed at a location
            z_dim         ::   The dimension for depth levels.
            var_user_want ::   List of observations the user wants to reshape       
        """

        # find maximum z levels in any of the profiles
        d_max = int(np.max(self.dataset.z_row_size.values))
        # number of profiles
        prof_size = self.dataset.z_row_size.shape[0]

        # set a 2D array (relevant to maximum depth)
        depth_2d = np.empty(
            (
                prof_size,
                d_max,
            )
        )
        depth_2d[:] = np.nan
        # reshape depth information from 1D to 2D
        if np.isnan(self.dataset.z_row_size.values[0]) == False:
            I_SIZE = int(self.dataset.z_row_size.values[0])
            depth_2d[0, :I_SIZE] = self.dataset.depth[0:I_SIZE].values
        for iJ in range(1, prof_size):
            if np.isnan(self.dataset.z_row_size.values[iJ]) == False:
                I_START = int(np.nansum(self.dataset.z_row_size.values[:iJ]))
                I_END = int(np.nansum(self.dataset.z_row_size.values[: iJ + 1]))
                I_SIZE = int(self.dataset.z_row_size.values[iJ])
                depth_2d[iJ, 0:I_SIZE] = self.dataset.depth[I_START:I_END].values

        # check reshape
        T_OBS1 = np.delete(np.ravel(depth_2d), np.isnan(np.ravel(depth_2d)))
        T_OBS2 = np.delete(self.dataset.depth.values, np.isnan(self.dataset.depth.values))
        if T_OBS1.size == T_OBS2.size and (int(np.min(T_OBS1 - T_OBS2)) == 0 or int(np.max(T_OBS1 - T_OBS2)) == 0):
            print("Depth OK reshape successful")
        else:
            print("Depth WRONG!! reshape")

        # reshape obs for each variable from 1D to 2D
        var_all = np.empty(
            (
                len(var_user_want),
                prof_size,
                d_max,
            )
        )
        var_list = var_user_want[:]
        counter_i = 0
        for iN in range(0, len(var_user_want)):
            print(var_user_want[iN])
            # check that variable exist in the WOD observations file
            if var_user_want[iN] in self.dataset:
                print("observed variable exist")
                # reshape it into 2D
                var_2d = np.empty(
                    (
                        prof_size,
                        d_max,
                    )
                )
                var_2d[:] = np.nan
                # populate array but make sure that the indexing for number of levels
                # is not nan, as in the data there are nan indexings for number of levels
                # indicating no observations there
                if np.isnan(self.dataset[var_user_want[iN] + "_row_size"][0].values) == False:
                    I_SIZE = int(self.dataset[var_user_want[iN] + "_row_size"][0].values)
                    var_2d[0, :I_SIZE] = self.dataset[var_user_want[iN]][0:I_SIZE].values
                for iJ in range(1, prof_size):
                    if np.isnan(self.dataset[var_user_want[iN] + "_row_size"].values[iJ]) == False:
                        I_START = int(np.nansum(self.dataset[var_user_want[iN] + "_row_size"].values[:iJ]))
                        I_END = int(np.nansum(self.dataset[var_user_want[iN] + "_row_size"].values[: iJ + 1]))
                        I_SIZE = int(self.dataset[var_user_want[iN] + "_row_size"].values[iJ])
                        var_2d[iJ, 0:I_SIZE] = self.dataset[var_user_want[iN]].values[I_START:I_END]

                # all variables in one array
                var_all[counter_i, :, :] = var_2d
                counter_i = counter_i + 1
                # check that you did everything correctly and the obs in yoru reshaped
                # array match the observations in original array
                T_OBS1 = np.delete(np.ravel(var_2d), np.isnan(np.ravel(var_2d)))
                del I_START, I_END, I_SIZE, var_2d
                T_OBS2 = np.delete(
                    self.dataset[var_user_want[iN]].values, np.isnan(self.dataset[var_user_want[iN]].values)
                )

                if T_OBS1.size == T_OBS2.size and (
                    int(np.min(T_OBS1 - T_OBS2)) == 0 or int(np.max(T_OBS1 - T_OBS2)) == 0
                ):
                    print("OK reshape successful")
                else:
                    print("WRONG!! reshape")

            else:
                print("variable not in observations")
                var_list[iN] = "NO"

        # REMOVE DUBLICATES
        var_list = list(dict.fromkeys(var_list))
        # REMOVE the non-observed variables from the list of variables
        var_list.remove("NO")

        # create the new 2D dataset array
        wod_profiles_2d = xr.Dataset(
            {
                "depth": (["id_dim", "z_dim"], depth_2d),
            },
            coords={
                "time": (["id_dim"], self.dataset.time.values),
                "latitude": (["id_dim"], self.dataset.latitude.values),
                "longitude": (["id_dim"], self.dataset.longitude.values),
            },
        )
        for iN in range(0, len(var_list)):
            wod_profiles_2d[var_list[iN]] = (["id_dim", "z_dim"], var_all[iN, :, :])

        return_prof = Profile()
        return_prof.dataset = wod_profiles_2d
        return return_prof

    def time_slice(self, date0, date1):
        """Return new Gridded object, indexed between dates date0 and date1"""
        dataset = self.dataset
        t_ind = pd.to_datetime(dataset.time.values) >= date0
        dataset = dataset.isel(id_dim=t_ind)
        t_ind = pd.to_datetime(dataset.time.values) < date1
        dataset = dataset.isel(id_dim=t_ind)
        return Profile(dataset=dataset)



class DocsyTools:  # TODO All abstract methods should be implemented
    """ """

    def __init__(self):
        return  # TODO Super __init__ should be called at some point

    @classmethod
    def write_class_to_markdown(
        cls, class_to_write, fn_out, method_to_omit=[], omit_private_methods=True, omit_parent_methods=True
    ):

        methods_to_write = cls._get_list_of_methods(class_to_write)

        for method in methods_to_write:
            method_block = cls._method_to_str(getattr(class_to_write, method))

    @classmethod
    def _method_to_str(cls, method_name):
        pass

    @classmethod
    def _get_list_of_methods(
        cls, class_to_search, methods_to_omit=[], omit_private_methods=True, omit_parent_methods=True
    ):
        """
        Returns a list of methods inside a provided COAsT class, with some
        other options

        Parameters
        ----------
        class_to_search : imported class
            Class imported from COAsT (e.g. from coast import Profile)
        methods_to_omit : list
            List of method strings to omit from the output. The default is [].
        omit_private_methods : bool, optional
            If true, omit methods beginning with "_". The default is True.
        omit_parent_methods : bool, optional
           If true, omit methods in any parent/ancestor class. The default is True.

        Returns
        -------
        methods_to_write : list
            List of strings denoting method names

        """

        # Get list of methods
        methods_to_write = dir(class_to_search)

        if omit_parent_methods:
            # Get parent class and methods (if user wants)
            parents = class_to_search.__mro__[1:-1]
            parent_methods = []

            # Get (non-unique) parent methods
            for pp in parents:
                methods_pp = dir(pp)
                [parent_methods.append(mm) for mm in methods_pp]

            # Remove classes from methods list if they are the same name
            for method in parent_methods:
                if method in methods_to_write:
                    methods_to_write.remove(method)

        # Look for methods beginning with "_" and remove them (if true)
        if omit_private_methods:
            for method in methods_to_write:
                if method[0] == "_":
                    methods_to_write.remove(method)

        return methods_to_write



"""Argos class"""
from .index import Indexed
import numpy as np
import xarray as xr
import pandas as pd
from .logging_util import get_slug, debug
from typing import Union
from pathlib import Path


class Argos(Indexed):
    """Class for reading Argos CSV formatted data files into an xarray object"""

    def __init__(self, file_path: str = None, config: Union[Path, str] = None):
        """Init Argos data object

        Args:
            file_path (str) : path of data fil
            config (Path or str) : configuration file
        """
        debug(f"Creating a new {get_slug(self)}")
        super().__init__(config)
        self.read_data(file_path)
        self.apply_config_mappings()
        print(f"{get_slug(self)} initialised")

    def read_data(self, file_path: str) -> None:
        """Read the data file

        Expected format and variable names are

            DATIM,LAT,LON,SST,SST_F,PSST,PSST_F,PPRES,PPRES_F,BEN

        xarray dataset to have dimension as time and coordinates as time, latitude and longitude

        Args:
            file_path (str) : path of data file
        """
        df = pd.read_csv(file_path, na_values=" ")

        self.dataset = xr.Dataset()
        var_names = list(df.columns)
        for name in var_names:
            if name == "DATIM":
                time = np.array(pd.to_datetime(df[name]))
                self.dataset[name] = xr.DataArray(list(time), dims=["time"])
                self.dataset = self.dataset.set_coords(name)
            else:
                self.dataset[name] = xr.DataArray(list(np.array(df[name])), dims=["time"])



"""Index class."""
from dask import array
from dask.distributed import Client
from .logging_util import get_slug, debug, info, warn, warning
from typing import Union
from pathlib import Path
from ast import literal_eval
from .coast import Coast
from .config_parser import ConfigParser


def setup_dask_client(workers: int = 2, threads: int = 2, memory_limit_per_worker: str = "2GB"):
    Client(n_workers=workers, threads_per_worker=threads, memory_limit=memory_limit_per_worker)


class Indexed(Coast):
    def __init__(self, config: Union[Path, str] = None):
        """Configuration init.

        Args:
            config (Union[Path, str]): path to json config file.
        """
        debug(f"Creating a new {get_slug(self)}")

        self.dataset = None
        self.chunks = None
        self.var_mapping = None
        self.dim_mapping = None
        self.coord_vars = None
        self.keep_all_vars = False

        if config:
            print(config)
            self.json_config = ConfigParser(config)
            self.chunks = self.json_config.config.chunks
            self.dim_mapping = self.json_config.config.dataset.dimension_map
            self.var_mapping = self.json_config.config.dataset.variable_map
            self.coord_vars = self.json_config.config.dataset.coord_var
            self.keep_all_vars = literal_eval(self.json_config.config.dataset.keep_all_vars)

    def apply_config_mappings(self) -> None:
        """Applies json configuration and mappings"""
        # We iterate through each mapping one by one which enables us to catch those variables that do not exist

        if self.dim_mapping is not None:
            for k in self.dim_mapping:
                try:
                    self.dataset = self.dataset.rename_dims({k: self.dim_mapping[k]})
                except ValueError as e:
                    debug(f"Warning: {str(e)}")

        if self.var_mapping is not None:
            keep_vars = []
            for k in self.var_mapping:
                try:
                    self.dataset = self.dataset.rename_vars({k: self.var_mapping[k]})
                    keep_vars.append(self.var_mapping[k])
                except ValueError as e:
                    debug(f"Warning: {str(e)}")

            if not self.keep_all_vars:
                self.dataset = self.dataset[keep_vars]

        if self.coord_vars is not None:
            self.dataset = self.dataset.set_coords(self.coord_vars)

    def insert_dataset(self, dataset, apply_config_mappings=False):
        """Insert a dataset straight into this object instance"""
        self.dataset = dataset
        if apply_config_mappings:
            self.apply_config_mappings()



"""Config parser."""
import json
from pathlib import Path
from typing import Union

from .config_structure import ConfigTypes, ConfigKeys, GriddedConfig, IndexedConfig, Dataset, Domain, CodeProcessing


class ConfigParser:
    """Class for parsing gridded and indexed configuration files."""

    def __init__(self, json_path: Union[Path, str]):
        """Config parser constructor.

        Args:
            json_path (Union[Path, str]): path to json config file.
        """
        with open(json_path, "r") as j:
            json_content = json.loads(j.read())
            conf_type = ConfigTypes(json_content[ConfigKeys.TYPE])
            if conf_type == ConfigTypes.GRIDDED:
                self.config = ConfigParser._parse_gridded(json_content)
            elif conf_type == ConfigTypes.INDEXED:
                self.config = ConfigParser._parse_indexed(json_content)

    @staticmethod
    def _parse_gridded(json_content: dict) -> GriddedConfig:
        """Static method to parse Gridded config files.

        Args:
            json_content (dict): Config file json.
        """
        dimensionality = json_content[ConfigKeys.DIMENSIONALITY]
        grid_ref = json_content[ConfigKeys.GRID_REF]
        proc_flags = json_content[ConfigKeys.PROC_FLAGS]
        chunks = json_content[ConfigKeys.CHUNKS]
        dataset = ConfigParser._get_datafile_object(json_content, ConfigKeys.DATASET)
        static_variables = ConfigParser._get_code_processing_object(json_content)
        try:
            domain = ConfigParser._get_datafile_object(json_content, ConfigKeys.DOMAIN)
        except KeyError:
            domain = None
        return GriddedConfig(
            dimensionality=dimensionality,
            grid_ref=grid_ref,
            chunks=chunks,
            dataset=dataset,
            domain=domain,
            processing_flags=proc_flags,
            code_processing=static_variables,
        )

    @staticmethod
    def _parse_indexed(json_content: dict) -> IndexedConfig:
        """Static method to parse Indexed config files.

        Args:
            json_content (dict): Config file json.
        """
        dimensionality = json_content[ConfigKeys.DIMENSIONALITY]
        proc_flags = json_content[ConfigKeys.PROC_FLAGS]
        chunks = json_content[ConfigKeys.CHUNKS]
        dataset = ConfigParser._get_datafile_object(json_content, ConfigKeys.DATASET)
        return IndexedConfig(dimensionality=dimensionality, chunks=chunks, dataset=dataset, processing_flags=proc_flags)

    @staticmethod
    def _get_code_processing_object(json_content: dict) -> CodeProcessing:
        """Static method to convert static_variables configs into objects.

        Args:
            json_content (dict): Config file json.
        """
        dataset_json = json_content[ConfigKeys.CODE_PROCESSING]
        return CodeProcessing(
            delete_variables=dataset_json[ConfigKeys.DEL_VAR],
            not_grid_variables=dataset_json[ConfigKeys.NO_GR_VAR],
        )

    @staticmethod
    def _get_datafile_object(json_content: dict, data_file_type: str) -> Union[Dataset, Domain]:
        """Static method to convert dataset and domain configs into objects.

        Args:
            json_content (dict): Config file json.
            data_file_type (str): key of datafile type (dataset or domain).
        """
        dataset_json = json_content[data_file_type]
        dataset_var = dataset_json[ConfigKeys.VAR_MAP]
        dataset_dim = dataset_json[ConfigKeys.DIM_MAP]

        try:
            dataset_keep_all_vars = dataset_json[ConfigKeys.KEEP_ALL_VARS]
        except KeyError:
            dataset_keep_all_vars = "False"

        if data_file_type is ConfigKeys.DATASET:
            dataset_coord_vars = dataset_json[ConfigKeys.COO_VAR]
            return Dataset(
                variable_map=dataset_var,
                dimension_map=dataset_dim,
                coord_var=dataset_coord_vars,
                keep_all_vars=dataset_keep_all_vars,
            )
        elif data_file_type is ConfigKeys.DOMAIN:
            return Domain(variable_map=dataset_var, dimension_map=dataset_dim, keep_all_vars=False)



"""Oceanparcels class for reading ocean parcels data."""
from .lagrangian import Lagrangian
import xarray as xr
from .logging_util import get_slug, debug, info, warn, warning
from typing import Union
from pathlib import Path


class Oceanparcels(Lagrangian):
    """Reading ocean parcels data (netcdf format) into an xarray object."""

    def __init__(self, file_path: str = None, config: Union[Path, str] = None):
        """Initialization and file reading.

        Args:
           file_path (str): path to data file
           config (Union[Path, str]): path to json config file.
        """
        debug(f"Creating a new {get_slug(self)}")
        super().__init__(config)

        if file_path is None:
            warn("Object created but no file or directory specified: \n" "{0}".format(str(self)), UserWarning)
        else:
            self.load_single(file_path)
            self.apply_config_mappings()

        print(f"{get_slug(self)} initialised")

    def load_single(self, file_path) -> None:
        """Loads a single file into object's dataset variable.

        Args:
            file_path (str): path to data file
        """
        self.dataset = xr.open_dataset(file_path, chunks=self.chunks)



from .coast import Coast
from .gridded import Gridded
from scipy.ndimage import convolve1d
from scipy import interpolate
import gsw
import os.path as path_lib
import xarray as xr
import numpy as np
from scipy.integrate import cumtrapz
import warnings
import traceback
from .logging_util import get_slug, debug, warn, info

# =============================================================================
# The TRANSECT module is a place for code related to transects only
# =============================================================================


class Transect:
    gravity = 9.8  # m s^-2
    earth_rotation_rate = 7.2921 * 10 ** (-5)  # rad/s

    @staticmethod
    def moving_average(array_to_smooth, window=2, axis=-1):
        """
        Returns the input array smoothed along the given axis using convolusion
        """
        debug(f"Fetching moving average for {array_to_smooth}")
        return convolve1d(array_to_smooth, np.ones(window), axis=axis) / window

    @staticmethod
    def interpolate_slice(variable_slice, depth, interpolated_depth=None):
        """
        Linearly interpolates the variable at a single time along the z_dim, which must be the
        first axis.

        Parameters
        ----------
        variable_slice : Variable to interpolate (z_dim, r_dim)
        depth : The depth at each z point for each point along the transect
        interpolated_depth : (optional) desired depth profile to interpolate to. If not supplied
            a uniform depth profile uniformaly spaced between zero and variable max depth will be used
            with a spacing of 2 metres.

        Returns
        -------
        interpolated_depth_variable_slice : Interpolated variable
        interpolated_depth : Interpolation depth

        """
        debug(f"Interpolating slice {variable_slice} at depths {depth}")
        if interpolated_depth is None:
            interpolated_depth = np.arange(0, np.nanmax(depth), 2)

        interpolated_depth_variable_slice = np.zeros((len(interpolated_depth), variable_slice.shape[-1]))
        for i in np.arange(0, variable_slice.shape[-1]):
            depth_func = interpolate.interp1d(depth[:, i], variable_slice[:, i], axis=0, bounds_error=False)
            interpolated_depth_variable_slice[:, i] = depth_func(interpolated_depth)

        return interpolated_depth_variable_slice, interpolated_depth

    @staticmethod
    def gen_z_levels(max_depth):
        """Generates a pre-defined 1d vertical depth coordinate,
        i.e. horizontal z-level vertical coordinates up to a supplied
        maximum depth, 'max_depth'

        Parameters
        ----------
        max_depth : int, bottom level depth
        """

        max_depth = max_depth + 650
        z_levels_0_50 = np.arange(0, 55, 5)
        z_levels_60_290 = np.arange(60, 300, 10)
        z_levels_300_600 = np.arange(300, 650, 50)
        z_levels_650_ = np.arange(650, max_depth + 150, 150)
        z_levels = np.concatenate((z_levels_0_50, z_levels_60_290, z_levels_300_600, z_levels_650_))
        z_levels = z_levels[z_levels <= max_depth]
        return z_levels

    def __init__(self, gridded: Coast, point_a: tuple = None, point_b: tuple = None, y_indices=None, x_indices=None):
        """
        Class defining a generic transect type, which is a 3d dataset along
        a linear path between a point A and a point B, with a time dimension,
        a depth dimension and an along transect dimension.
        The model Data on the supplied grid is subsetted in its entirety along these dimensions.

        Note that Point A should be closer to the southern boundary of the model domain.

        The user can either supply the start and end (lat,lon) coordinates of the
        transect, point_A and point_B respectively, or the model y, x indices defining it.
        In the latter case the user must ensure that the indices define a continuous
        transect, e.g. y=[10,11,11,12], x=[5,5,6,6].
        Only limited checks are performed on the suitability of the indices.

        Example usage:
            point_A = (54,-15)
            point_B = (56,-12)
            transect = coast.Transect( gridded_t, point_A, point_B )
            or
            transect = coast.Transect( gridded_f, y_indices=y_ind, x_indices=x_ind )

        Parameters
        ----------
        gridded : GRIDDED object
        point_a : tuple, (lat,lon)
        point_b : tuple, (lat,lon)
        y_indices : 1d array of model y indices defining the points of the transect
        x_indices : 1d array of model x indices defining the points of the transect

        """
        debug(f"Creating a new {get_slug(self)}")
        try:
            self.filename_domain = gridded.filename_domain

            if point_a is not None and point_b is not None:
                # point A should be of lower latitude than point B
                if abs(point_b[0]) < abs(point_a[0]):
                    self.point_A = point_b
                    self.point_B = point_a
                else:
                    self.point_A = point_a
                    self.point_B = point_b

                # Get points on transect
                tran_y_ind, tran_x_ind, tran_len = gridded.transect_indices(self.point_A, self.point_B)
                tran_y_ind, tran_x_ind = self.process_transect_indices(
                    gridded, np.asarray(tran_y_ind), np.asarray(tran_x_ind)
                )
            elif y_indices is not None and x_indices is not None:
                if y_indices[0] > y_indices[-1]:
                    y_indices = y_indices[::-1]
                    x_indices = x_indices[::-1]
                tran_y_ind, tran_x_ind = self.process_transect_indices(gridded, y_indices, x_indices)
                self.point_A = (
                    gridded.dataset.latitude[tran_y_ind[0], tran_x_ind[0]],
                    gridded.dataset.longitude[tran_y_ind[0], tran_x_ind[0]],
                )
                self.point_B = (
                    gridded.dataset.latitude[tran_y_ind[-1], tran_x_ind[-1]],
                    gridded.dataset.longitude[tran_y_ind[-1], tran_x_ind[-1]],
                )
            else:
                raise ValueError(
                    "Must supply both point_A and point_B of transect \
                                 or the indices defining it."
                )

            # indices along the transect
            self.y_ind = tran_y_ind
            self.x_ind = tran_x_ind
            self.len = len(tran_y_ind)
            self.data_cross_tran_flow = xr.Dataset()

            # Subset the gridded data along the transect creating a new dimension (r_dim),
            # which is a paramterisation for x_dim and y_dim defining the transect
            da_tran_y_ind = xr.DataArray(tran_y_ind, dims=["r_dim"])
            da_tran_x_ind = xr.DataArray(tran_x_ind, dims=["r_dim"])
            self.data = gridded.dataset.isel(y_dim=da_tran_y_ind, x_dim=da_tran_x_ind)

            debug(f"{get_slug(self)} initialised")
        except ValueError:
            print(traceback.format_exc())

    def process_transect_indices(self, gridded, tran_y_ind, tran_x_ind):
        """
        Get the transect indices on a specific grid

        Parameters
        ----------
        gridded : the model grid to define the transect on

        Return
        ----------
        tran_y_ind : array of y_dim indices
        tran_x_ind : array of x_dim indices

        """
        debug(f"Fetching transect indices for {get_slug(self)} with {get_slug(gridded)}")
        try:
            # Redefine transect so that each point on the transect is seperated
            # from its neighbours by a single index change in y or x, but not both
            dist_option_1 = (
                gridded.dataset.e2.values[tran_y_ind, tran_x_ind]
                + gridded.dataset.e1.values[tran_y_ind + 1, tran_x_ind]
            )
            dist_option_2 = (
                gridded.dataset.e2.values[tran_y_ind, tran_x_ind + 1]
                + gridded.dataset.e1.values[tran_y_ind, tran_x_ind]
            )
            spacing = np.abs(np.diff(tran_y_ind)) + np.abs(np.diff(tran_x_ind))
            if spacing.max() > 2:
                raise ValueError(
                    "The transect is not continuous. The transect must be defined on " "adjacent grid points."
                )
            spacing[spacing != 2] = 0
            double_spacing = np.nonzero(spacing)[0]
            for space in double_spacing[::-1]:
                if dist_option_1[space] < dist_option_2[space]:
                    tran_y_ind = np.insert(tran_y_ind, space + 1, tran_y_ind[space + 1])
                    tran_x_ind = np.insert(tran_x_ind, space + 1, tran_x_ind[space])
                else:
                    tran_y_ind = np.insert(tran_y_ind, space + 1, tran_y_ind[space])
                    tran_x_ind = np.insert(tran_x_ind, space + 1, tran_x_ind[space + 1])
            return tran_y_ind, tran_x_ind
        except ValueError:
            print(traceback.format_exc())

    def plot_transect_on_map(self):
        """
        Plot transect location on a map

        Example usage:
        --------------
        tran = coast.Transect( (54,-15), (56,-12), gridded )
        tran.plot_map()
        """
        debug(f"Generating plot on map for {get_slug(self)}")
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys

            warnings.warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(10, 10))
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        ax.add_feature(cartopy.feature.BORDERS, linestyle=":")
        coast = NaturalEarthFeature(
            category="physical", scale="50m", facecolor=[0.8, 0.8, 0.8], name="coastline", alpha=0.5
        )
        ax.add_feature(coast, edgecolor="gray")

        gl = ax.gridlines(
            crs=ccrs.PlateCarree(), draw_labels=True, linewidth=0.5, color="gray", alpha=0.5, linestyle="-"
        )

        gl.top_labels = False
        gl.bottom_labels = True
        gl.right_labels = False
        gl.left_labels = True
        gl.x_formatter = LONGITUDE_FORMATTER
        gl.y_formatter = LATITUDE_FORMATTER

        plt.title("Map of transect location")
        # plt.show() # Can only adjust axis if fig is plotted already..

        return fig, ax


class TransectF(Transect):
    """
    Class defining a transect on the f-grid, which is a 3d dataset along
    a linear path between a point A and a point B, with a time dimension,
    a depth dimension and an along transect dimension. The model Data on f-grid
    is subsetted in its entirety along these dimensions.

    Note that Point A should be closer to the southern boundary of the model domain.

    The user can either supply the start and end (lat,lon) coordinates of the
    transect, point_A and point_B respectively, or the model y, x indices defining it.
    In the latter case the user must ensure that the indices define a continuous
    transect, e.g. y=[10,11,11,12], x=[5,5,6,6].
    Only limited checks are performed on the suitability of the indices.

    Example usage:
        point_A = (54,-15)
        point_B = (56,-12)
        transect = coast.Transect_f( gridded_f, point_A, point_B )
        or
        transect = coast.Transect_f( gridded_f, y_indices=y_ind, x_indices=x_ind )

    Parameters
    ----------
    gridded_f : GRIDDED object on the f-grid
    point_a : tuple, (lat,lon)
    point_b : tuple, (lat,lon)
    y_indices : 1d array of model y indices defining the points of the transect
    x_indices : 1d array of model x indices defining the points of the transect

    """

    def __init__(self, gridded_f: Coast, point_a: tuple = None, point_b: tuple = None, y_indices=None, x_indices=None):
        super().__init__(gridded_f, point_a, point_b, y_indices, x_indices)

    def calc_flow_across_transect(self, gridded_u: Coast, gridded_v: Coast):
        # TODO the code here has become a little messy, could do with being
        # rewritten as in similar function in CONTOUR
        """

        Computes the flow through the transect at each segment and creates a new
        dataset 'Transect_f.data_cross_tran_flow' defined on the normal velocity
        points along the transect.
        Transect normal velocities are calculated at each grid point and stored in
        in Transect_f.data_cross_tran_flow.normal_velocities,
        Depth integrated volume transport across the transect is calculated
        at each transect segment and stored in Transect_f.data_cross_tran_flow.normal_transports
        The latitude, longitude and the horizontal and vertical scale factors
        on the normal velocity points are also stored in the dataset.

        If the time dependent cell thicknesses (e3) on the u and v grids are
        present in the gridded_u and gridded_v datasets they will be used, if they
        are not then the initial cell thicknesses (e3_0) will be used.

        parameters
        ----------
        gridded_u : GRIDDED object on the u-grid containing the i-component velocities
        gridded_v : GRIDDED object on the v-gridc ontaining the j-component velocities

        """
        debug(f"Computing flow across the transect for {get_slug(self)}")

        # compute transports flag; set to false if suitable e3 not found
        compute_transports = True

        # subset the u and v datasets
        da_y_ind = xr.DataArray(self.y_ind, dims=["r_dim"])
        da_x_ind = xr.DataArray(self.x_ind, dims=["r_dim"])
        u_ds = gridded_u.dataset.isel(y_dim=da_y_ind, x_dim=da_x_ind)
        v_ds = gridded_v.dataset.isel(y_dim=da_y_ind, x_dim=da_x_ind)

        # use time varying if e3 is present, if not default to e3_0
        if "e3" not in u_ds.data_vars:
            if "e3_0" not in u_ds.data_vars:
                warn("e3 not found, transports will not be calculated")
                compute_transports = False
            else:
                u_ds["e3"] = u_ds.e3_0.broadcast_like(u_ds.u_velocity)
        if "e3" not in v_ds.data_vars:
            if "e3_0" not in v_ds.data_vars:
                warn("e3 not found, transports will not be calculated")
                compute_transports = False
            else:
                v_ds["e3"] = v_ds.e3_0.broadcast_like(v_ds.v_velocity)

        # If there is no time dimension, add one. This is so
        # indexing can assume a time dimension exists
        droptime = False
        if "t_dim" not in u_ds.dims:
            u_ds["u_velocity"] = u_ds.u_velocity.expand_dims(dim={"t_dim": 1}, axis=0)
            if compute_transports:
                u_ds["e3"] = u_ds.e3.expand_dims(dim={"t_dim": 1}, axis=0)
            droptime = True
        if "t_dim" not in v_ds.dims:
            v_ds["v_velocity"] = v_ds.v_velocity.expand_dims(dim={"t_dim": 1}, axis=0)
            if compute_transports:
                v_ds["e3"] = v_ds.e3.expand_dims(dim={"t_dim": 1}, axis=0)

        velocity = np.ma.zeros((u_ds.t_dim.size, u_ds.z_dim.size, u_ds.r_dim.size - 1))
        vol_transport = np.ma.zeros((u_ds.t_dim.size, u_ds.z_dim.size, u_ds.r_dim.size - 1))
        depth_0 = np.ma.zeros((u_ds.z_dim.size, u_ds.r_dim.size - 1))
        latitude = np.ma.zeros((u_ds.r_dim.size - 1))
        longitude = np.ma.zeros((u_ds.r_dim.size - 1))
        e1 = np.ma.zeros((u_ds.r_dim.size - 1))
        e2 = np.ma.zeros((u_ds.r_dim.size - 1))
        e3 = np.ma.zeros((u_ds.t_dim.size, u_ds.z_dim.size, u_ds.r_dim.size - 1))

        # Find the indices where the derivative of the transect in the north, south, east and west
        # directions are positive.
        dr_n = np.where(np.diff(self.y_ind) > 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_n = dr_n[~np.isnan(dr_n)].astype(int)
        dr_e = np.where(np.diff(self.x_ind) > 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_e = dr_e[~np.isnan(dr_e)].astype(int)
        dr_w = np.where(np.diff(self.x_ind) < 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_w = dr_w[~np.isnan(dr_w)].astype(int)

        # u flux (+ in)
        velocity[:, :, dr_n] = u_ds.u_velocity.to_masked_array()[:, :, dr_n + 1]
        if compute_transports:
            vol_transport[:, :, dr_n] = (
                velocity[:, :, dr_n] * u_ds.e2.to_masked_array()[dr_n + 1] * u_ds.e3.to_masked_array()[:, :, dr_n + 1]
            )
            e3[:, :, dr_n] = u_ds.e3.values[:, :, dr_n + 1]
        depth_0[:, dr_n] = u_ds.depth_0.to_masked_array()[:, dr_n + 1]
        latitude[dr_n] = u_ds.latitude.values[dr_n + 1]
        longitude[dr_n] = u_ds.longitude.values[dr_n + 1]
        e1[dr_n] = u_ds.e1.values[dr_n + 1]
        e2[dr_n] = u_ds.e2.values[dr_n + 1]

        # v flux (- in)
        velocity[:, :, dr_e] = -v_ds.v_velocity.to_masked_array()[:, :, dr_e + 1]
        if compute_transports:
            vol_transport[:, :, dr_e] = (
                velocity[:, :, dr_e] * v_ds.e1.to_masked_array()[dr_e + 1] * v_ds.e3.to_masked_array()[:, :, dr_e + 1]
            )
            e3[:, :, dr_e] = v_ds.e3.values[:, :, dr_e + 1]
        depth_0[:, dr_e] = v_ds.depth_0.to_masked_array()[:, dr_e + 1]
        latitude[dr_e] = v_ds.latitude.values[dr_e + 1]
        longitude[dr_e] = v_ds.longitude.values[dr_e + 1]
        e1[dr_e] = v_ds.e1.values[dr_e + 1]
        e2[dr_e] = v_ds.e2.values[dr_e + 1]

        # v flux (+ in)
        velocity[:, :, dr_w] = v_ds.v_velocity.to_masked_array()[:, :, dr_w]
        if compute_transports:
            vol_transport[:, :, dr_w] = (
                velocity[:, :, dr_w] * v_ds.e1.to_masked_array()[dr_w] * v_ds.e3.to_masked_array()[:, :, dr_w]
            )
            e3[:, :, dr_w] = v_ds.e3.values[:, :, dr_w]
        depth_0[:, dr_w] = v_ds.depth_0.to_masked_array()[:, dr_w]
        latitude[dr_w] = v_ds.latitude.values[dr_w]
        longitude[dr_w] = v_ds.longitude.values[dr_w]
        e1[dr_w] = v_ds.e1.values[dr_w]
        e2[dr_w] = v_ds.e2.values[dr_w]

        # Add DataArrays to dataset
        if droptime:
            self.data_cross_tran_flow["normal_velocities"] = xr.DataArray(
                velocity.squeeze(),
                coords={
                    "depth_0": (("z_dim", "r_dim"), depth_0),
                    "latitude": ("r_dim", latitude),
                    "longitude": ("r_dim", longitude),
                },
                dims=["z_dim", "r_dim"],
            )
            if compute_transports:
                self.data_cross_tran_flow["normal_transports"] = xr.DataArray(
                    np.sum(vol_transport.squeeze(), axis=0) / 1000000.0,
                    coords={"latitude": ("r_dim", latitude), "longitude": ("r_dim", longitude)},
                    dims=["r_dim"],
                ).squeeze()
        else:
            self.data_cross_tran_flow["normal_velocities"] = xr.DataArray(
                velocity,
                coords={
                    "time": ("t_dim", u_ds.time.values),
                    "depth_0": (("z_dim", "r_dim"), depth_0),
                    "latitude": ("r_dim", latitude),
                    "longitude": ("r_dim", longitude),
                },
                dims=["t_dim", "z_dim", "r_dim"],
            )
            if compute_transports:
                self.data_cross_tran_flow["normal_transports"] = xr.DataArray(
                    np.sum(vol_transport, axis=1) / 1000000.0,
                    coords={
                        "time": ("t_dim", u_ds.time.values),
                        "latitude": ("r_dim", latitude),
                        "longitude": ("r_dim", longitude),
                    },
                    dims=["t_dim", "r_dim"],
                ).squeeze()
        self.data_cross_tran_flow["e1"] = xr.DataArray(e1, dims=["r_dim"])
        self.data_cross_tran_flow["e2"] = xr.DataArray(e2, dims=["r_dim"])
        if compute_transports:
            self.data_cross_tran_flow["e3"] = xr.DataArray(e3, dims=["t_dim", "z_dim", "r_dim"])
        # DataArray attributes
        self.data_cross_tran_flow.normal_velocities.attrs["units"] = "m/s"
        self.data_cross_tran_flow.normal_velocities.attrs["standard_name"] = "velocity across the transect"
        self.data_cross_tran_flow.normal_velocities.attrs[
            "long_name"
        ] = "velocity across the transect defined on the normal velocity grid points"
        if compute_transports:
            self.data_cross_tran_flow.normal_transports.attrs["units"] = "Sv"
            self.data_cross_tran_flow.normal_transports.attrs[
                "standard_name"
            ] = "depth integrated volume transport across transect"
            self.data_cross_tran_flow.normal_transports.attrs[
                "long_name"
            ] = "depth integrated volume transport across the transect defined on the normal velocity grid points"
        self.data_cross_tran_flow.depth_0.attrs["units"] = "m"
        self.data_cross_tran_flow.depth_0.attrs["standard_name"] = "depth"
        self.data_cross_tran_flow.depth_0.attrs[
            "long_name"
        ] = "Initial depth at time zero defined at the normal velocity grid points"
        self.data_cross_tran_flow = self.data_cross_tran_flow.squeeze()

    @staticmethod
    def _pressure_grad_fpoint(ds_t, ds_t_j1, ds_t_i1, ds_t_j1i1, r_ind, velocity_component):
        """
        Calculates the hydrostatic and surface pressure gradients at a set of f-points
        along the transect, i.e. at a set of specific values of r_dim (but for all time and depth).
        The caller must supply four datasets that contain the variables which define
        the hydrostatic and surface pressure at all vertical z_levels and all time
        on the t-points around the transect i.e. for a set of f-points on the transect
        defined at (j+1/2, i+1/2), t-points are supplied at (j,i), (j+1,i), (j,i+1), (j+1,i+1),
        corresponding to ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1, respectively.

        The velocity_component defines whether u or v is normal to the transect
        for the segments of the transect. A segment of transect is
        defined as being r_dim to r_dim+1 where r_dim is the along transect dimension.

        Parameters
        ----------
        ds_t : coast.Transect_t on y=self.y_ind, x=self.x_ind
        ds_t_j1 : coast.Transect_t on y=self.y_ind+1, x=self.x_ind
        ds_t_i1 : coast.Transect_t on y=self.y_ind, x=self.x_ind+1
        ds_t_j1i1 : coast.Transect_t on y=self.y_ind+1, x=self.x_ind+1
        r_ind: 1d array, along transect indices
        velocity_component : str, normal velocity at r_ind

        Returns
        -------
        hpg_f : DataArray with dimensions in time and depth and along transect
            hydrostatic pressure gradient at a set of f-points along the transect
            for all time and depth
        spg_f : DataArray with dimensions in time and depth and along transect
            surface pressure gradient at a set of f-points along the transect

        """
        if velocity_component == "u":
            # required scale factors for derivative and averaging
            e2v = 0.5 * (ds_t_j1.e2.data[r_ind] + ds_t.e2.data[r_ind])
            e2v_i1 = 0.5 * (ds_t_j1i1.e2.data[r_ind] + ds_t_i1.e2.data[r_ind])
            e1v = 0.5 * (ds_t_j1.e1.data[r_ind] + ds_t.e1.data[r_ind])
            e1v_i1 = 0.5 * (ds_t_j1i1.e1.data[r_ind] + ds_t_i1.e1.data[r_ind])
            e1f = 0.5 * (e1v + e1v_i1)
            # calculate gradients at v-points either side of f-point
            hpg = (ds_t_j1.pressure_h_zlevels.data[:, :, r_ind] - ds_t.pressure_h_zlevels.data[:, :, r_ind]) / e2v
            hpg_i1 = (
                ds_t_j1i1.pressure_h_zlevels.data[:, :, r_ind] - ds_t_i1.pressure_h_zlevels.data[:, :, r_ind]
            ) / e2v_i1
            # average onto f-point
            hpg_f = 0.5 * ((e1v * hpg) + (e1v_i1 * hpg_i1)) / e1f
            # as aboave
            spg = (ds_t_j1.pressure_s.data[:, r_ind] - ds_t.pressure_s.data[:, r_ind]) / e2v
            spg_i1 = (ds_t_j1i1.pressure_s.data[:, r_ind] - ds_t_i1.pressure_s.data[:, r_ind]) / e2v_i1
            spg_f = 0.5 * ((e1v * spg) + (e1v_i1 * spg_i1)) / e1f
        elif velocity_component == "v":
            # required scale factors for derivative and averaging
            e1u = 0.5 * (ds_t_i1.e1.data[r_ind] + ds_t.e1.data[r_ind])
            e1u_j1 = 0.5 * (ds_t_j1i1.e1.data[r_ind] + ds_t_j1.e1.data[r_ind])
            e2u = 0.5 * (ds_t_i1.e2.data[r_ind] + ds_t.e2.data[r_ind])
            e2u_j1 = 0.5 * (ds_t_j1i1.e2.data[r_ind] + ds_t_j1.e2.data[r_ind])
            e2f = 0.5 * (e2u + e2u_j1)
            # calculate gradients at u-points either side of f-point
            hpg = (ds_t_i1.pressure_h_zlevels.data[:, :, r_ind] - ds_t.pressure_h_zlevels.data[:, :, r_ind]) / e1u
            hpg_j1 = (
                ds_t_j1i1.pressure_h_zlevels.data[:, :, r_ind] - ds_t_j1.pressure_h_zlevels.data[:, :, r_ind]
            ) / e1u_j1
            # average onto f-point
            hpg_f = 0.5 * ((e2u * hpg) + (e2u_j1 * hpg_j1)) / e2f
            # as above
            spg = (ds_t_i1.pressure_s.data[:, r_ind] - ds_t.pressure_s.data[:, r_ind]) / e1u
            spg_j1 = (ds_t_j1i1.pressure_s.data[:, r_ind] - ds_t_j1.pressure_s.data[:, r_ind]) / e1u_j1
            spg_f = 0.5 * ((e2u * spg) + (e2u_j1 * spg_j1)) / e2f

        return hpg_f, spg_f

    def calc_geostrophic_flow(
        self,
        gridded_t: Coast,
        ref_density=None,
        config_u="config/example_nemo_grid_u.json",
        config_v="config/example_nemo_grid_v.json",
    ):
        """
        This method will calculate the geostrophic velocity and volume transport
        (due to the geostrophic current) across the transect.
        4 variables are added to the Transect_f.data_cross_tran_flow dataset:
            1. normal_velocity_hpg      (t_dim, depth_z_levels, r_dim)
            This is the velocity due to the hydrostatic pressure gradient
            2. normal_velocity_spg      (t_dim, r_dim)
            This is the velocity due to the surface pressure gradient
            3. normal_transport_hpg  (t_dim, r_dim)
            This is the volume transport due to the hydrostatic pressure gradient
            4. normal_transport_AB_spg  (t_dim, r_dim
            This is the volume transport due to the surface pressure gradient

        The implementation works by regridding from the native vertical grid to
        horizontal z_levels in order to perform the horizontal gradients.
        Currently the level depths are assumed fixed at their initial depths,
        i.e. at time zero.

        Parameters
        ----------
        gridded_t : Coast
            This is gridded model data on the t-grid for the entire domain. It
            must contain the temperature, salinity and t-grid domain data (e1t, e2t, e3t_0).
        ref_density : float, optional
            reference density value. If None a transect mean density will be calculated
            and used.
        config_u : file
            configuration file for u-grid object
        config_v : file
            configuration file for v-grid object

        """
        debug(
            f"Calculating geostrophic velocity and volume transport for {get_slug(self)} with " f"{get_slug(gridded_t)}"
        )

        # If there is no time dimension, add one then remove at end. This is so
        # indexing can assume a time dimension exists
        gridded_t_local = gridded_t.copy()
        if "t_dim" not in gridded_t_local.dataset.dims:
            gridded_t_local.dataset = gridded_t_local.dataset.expand_dims(dim={"t_dim": 1}, axis=0)

        # We need to calculate the pressure at four t-points to get an
        # average onto the pressure gradient at the f-points, which will then
        # be averaged onto the normal velocity points. Here we subset the gridded_t
        # data around the transect so we have these four t-grid points at each
        # point along the transect
        tran_t = TransectT(gridded_t_local, y_indices=self.y_ind, x_indices=self.x_ind)  # j,i
        tran_t_j1 = TransectT(gridded_t_local, y_indices=self.y_ind + 1, x_indices=self.x_ind)  # j+1,i
        tran_t_i1 = TransectT(gridded_t_local, y_indices=self.y_ind, x_indices=self.x_ind + 1)  # j,i+1
        tran_t_j1i1 = TransectT(gridded_t_local, y_indices=self.y_ind + 1, x_indices=self.x_ind + 1)  # j+1,i+1

        bath_max = np.max(
            [
                tran_t.data.bathymetry.max().item(),
                tran_t_j1.data.bathymetry.max().item(),
                tran_t_i1.data.bathymetry.max().item(),
                tran_t_j1i1.data.bathymetry.max().item(),
            ]
        )

        z_levels = Transect.gen_z_levels(bath_max)

        tran_t.construct_pressure(ref_density, z_levels, extrapolate=True)
        tran_t_j1.construct_pressure(ref_density, z_levels, extrapolate=True)
        tran_t_i1.construct_pressure(ref_density, z_levels, extrapolate=True)
        tran_t_j1i1.construct_pressure(ref_density, z_levels, extrapolate=True)

        # Remove the mean hydrostatic pressure on each z_level from the hydrostatic pressure.
        # This helps to reduce the noise when taking the horizontal gradients of hydrostatic pressure.
        # Also catch and ignore nan-slice warning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            pressure_h_zlevel_mean = xr.concat(
                (
                    tran_t.data.pressure_h_zlevels,
                    tran_t_j1.data.pressure_h_zlevels,
                    tran_t_i1.data.pressure_h_zlevels,
                    tran_t_j1i1.data.pressure_h_zlevels,
                ),
                dim="concat_dim",
            ).mean(dim=("concat_dim", "r_dim", "t_dim"), skipna=True)

            if ref_density is None:
                ref_density = (
                    xr.concat(
                        (
                            tran_t.data.density_zlevels,
                            tran_t_j1.data.density_zlevels,
                            tran_t_i1.data.density_zlevels,
                            tran_t_j1i1.data.density_zlevels,
                        ),
                        dim="concat_dim",
                    )
                    .mean(dim=("concat_dim", "r_dim", "t_dim", "depth_z_levels"), skipna=True)
                    .item()
                )

        tran_t.data["pressure_h_zlevels"] = tran_t.data.pressure_h_zlevels - pressure_h_zlevel_mean
        tran_t_j1.data["pressure_h_zlevels"] = tran_t_j1.data.pressure_h_zlevels - pressure_h_zlevel_mean
        tran_t_i1.data["pressure_h_zlevels"] = tran_t_i1.data.pressure_h_zlevels - pressure_h_zlevel_mean
        tran_t_j1i1.data["pressure_h_zlevels"] = tran_t_j1i1.data.pressure_h_zlevels - pressure_h_zlevel_mean

        # Coriolis parameter
        f = 2 * self.earth_rotation_rate * np.sin(np.deg2rad(self.data.latitude))

        # Find the indices where the derivative of the transect in the north, south, east and west
        # directions are positive.
        dr_n = np.where(np.diff(self.y_ind) > 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_e = np.where(np.diff(self.x_ind) > 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_w = np.where(np.diff(self.x_ind) < 0, np.arange(0, self.data.r_dim.size - 1), np.nan)
        dr_list = [
            dr_n[~np.isnan(dr_n)].astype(int),
            dr_e[~np.isnan(dr_e)].astype(int),
            dr_w[~np.isnan(dr_w)].astype(int),
        ]

        # horizontal scale factors on the relevent u and v grids that are
        # normal to the transect for dr_n, dr_e, dr_w
        e2u_j1 = 0.5 * (tran_t_j1.data.e2.data[dr_list[0]] + tran_t_j1i1.data.e2.data[dr_list[0]])
        e1v_i1 = 0.5 * (tran_t_i1.data.e1.data[dr_list[1]] + tran_t_j1i1.data.e1.data[dr_list[1]])
        e1v = 0.5 * (tran_t.data.e1.data[dr_list[2]] + tran_t_j1.data.e1.data[dr_list[2]])
        e_horiz_vel = [e2u_j1, e1v_i1, e1v]
        # Horizontal scale factors on f-grid for dr_n, dr_e, dr_w
        e_horiz_f = [self.data.e2, self.data.e1, self.data.e1]
        # velocity component normal to transect for dr_n, dr_s, dr_e, dr_w
        velocity_component = ["u", "v", "v"]
        # Geostrophic flow direction across transect
        flow_direction = [-1, -1, 1]

        # The cross transect flow is defined on the u and v points that are across
        # the transect, i.e. between f points, therefore the attributes of the
        # data_cross_flow dataset need to be on these points.
        da_y_ind = xr.DataArray(self.y_ind, dims=["r_dim"])
        da_x_ind = xr.DataArray(self.x_ind, dims=["r_dim"])
        u_ds = Gridded(fn_domain=self.filename_domain, config=config_u).dataset.isel(y_dim=da_y_ind, x_dim=da_x_ind)
        v_ds = Gridded(fn_domain=self.filename_domain, config=config_v).dataset.isel(y_dim=da_y_ind, x_dim=da_x_ind)
        ds = [u_ds, v_ds, v_ds]

        # Drop the last point because the normal velocity points are defined at
        # the middle of a segment and there is as a result one less point.
        normal_velocity_hpg = np.zeros_like(tran_t.data.pressure_h_zlevels)[:, :, :-1]
        normal_velocity_spg = np.zeros_like(tran_t.data.pressure_s)[:, :-1]
        latitude = np.zeros((u_ds.r_dim.size - 1))
        longitude = np.zeros((u_ds.r_dim.size - 1))
        depth_0 = np.ma.zeros((u_ds.z_dim.size, u_ds.r_dim.size - 1))
        # horizontal scale factors for each segmant of transect
        e_horiz = np.zeros((tran_t.data.t_dim.size, tran_t.data.r_dim.size - 1))
        # Contruct geostrophic flow
        for dr, vel_comp, flow_dir, e_hor_vel, e_hor_f, i_ds in zip(
            dr_list, velocity_component, flow_direction, e_horiz_vel, e_horiz_f, ds
        ):
            hpg, spg = self._pressure_grad_fpoint(
                tran_t.data, tran_t_j1.data, tran_t_i1.data, tran_t_j1i1.data, dr, vel_comp
            )
            hpg_r1, spg_r1 = self._pressure_grad_fpoint(
                tran_t.data, tran_t_j1.data, tran_t_i1.data, tran_t_j1i1.data, dr + 1, vel_comp
            )
            normal_velocity_hpg[:, :, dr] = (
                flow_dir
                * 0.5
                * (e_hor_f.data[dr] * hpg / f.data[dr] + e_hor_f.data[dr + 1] * hpg_r1 / f.data[dr + 1])
                / (e_hor_vel * ref_density)
            )
            normal_velocity_spg[:, dr] = (
                flow_dir
                * 0.5
                * (e_hor_f.data[dr] * spg / f.data[dr] + e_hor_f.data[dr + 1] * spg_r1 / f.data[dr + 1])
                / (e_hor_vel * ref_density)
            )
            e_horiz[:, dr] = e_hor_vel
            depth_0[:, dr] = i_ds.depth_0.to_masked_array()[:, dr]
            latitude[dr] = i_ds.latitude.data[dr]
            longitude[dr] = i_ds.longitude.data[dr]

        # Bathymetry at normal velocity points
        # H = np.zeros_like( self.data.bathymetry.values )[:-1]
        H = 0.5 * (self.data.bathymetry.values[:-1] + self.data.bathymetry.values[1:])
        # Remove redundent levels below bathymetry
        normal_velocity_hpg = np.where(z_levels[:, np.newaxis] <= H, normal_velocity_hpg, np.nan)
        active_z_levels = np.count_nonzero(~np.isnan(normal_velocity_hpg), axis=1).max()
        normal_velocity_hpg = normal_velocity_hpg[:, :active_z_levels, :]
        z_levels = z_levels[:active_z_levels]

        # DataArray attributes
        coords_hpg = {
            "depth_z_levels": ("depth_z_levels", z_levels),
            "latitude": ("r_dim", latitude),
            "longitude": ("r_dim", longitude),
        }
        dims_hpg = ["depth_z_levels", "r_dim"]
        attributes_hpg = {
            "units": "m/s",
            "standard name": "velocity across the \
                          transect due to the hydrostatic pressure gradient",
        }
        coords_spg = {"latitude": ("r_dim", latitude), "longitude": ("r_dim", longitude)}
        dims_spg = ["r_dim"]
        attributes_spg = {
            "units": "m/s",
            "standard name": "velocity across the \
                          transect due to the surface pressure gradient",
        }

        # Add time if required
        if "t_dim" in tran_t.data.dims:
            coords_hpg["time"] = ("t_dim", tran_t.data.time.values)
            dims_hpg.insert(0, "t_dim")
            coords_spg["time"] = ("t_dim", tran_t.data.time.values)
            dims_spg.insert(0, "t_dim")

        # Add DataArrays  to dataset
        self.data_cross_tran_flow["normal_velocity_hpg"] = xr.DataArray(
            np.squeeze(normal_velocity_hpg), coords=coords_hpg, dims=dims_hpg, attrs=attributes_hpg
        )
        self.data_cross_tran_flow["normal_velocity_spg"] = xr.DataArray(
            np.squeeze(normal_velocity_spg), coords=coords_spg, dims=dims_spg, attrs=attributes_spg
        )
        self.data_cross_tran_flow["normal_transport_hpg"] = (
            (self.data_cross_tran_flow.normal_velocity_hpg.fillna(0).integrate(coord="depth_z_levels"))
            * e_horiz
            / 1000000
        )
        self.data_cross_tran_flow.normal_transport_hpg.attrs = {
            "units": "Sv",
            "standard_name": "volume transport across transect due to the hydrostatic pressure gradient",
        }
        self.data_cross_tran_flow["normal_transport_spg"] = (
            self.data_cross_tran_flow.normal_velocity_spg * H * e_horiz / 1000000
        )
        self.data_cross_tran_flow.normal_transport_spg.attrs = {
            "units": "Sv",
            "standard_name": "volume transport across transect due to the surface pressure gradient",
        }

        self.data_cross_tran_flow["latitude"] = xr.DataArray(latitude, dims=["r_dim"])
        self.data_cross_tran_flow["longitude"] = xr.DataArray(longitude, dims=["r_dim"])
        self.data_cross_tran_flow["e12"] = xr.DataArray(e_horiz[0, :], dims=["r_dim"])
        self.data_cross_tran_flow["depth_0_original"] = xr.DataArray(depth_0, dims=["z_dim", "r_dim"])
        self.data_cross_tran_flow.depth_0_original.attrs["units"] = "m"
        self.data_cross_tran_flow.depth_0_original.attrs["standard_name"] = "original depth coordinate"
        self.data_cross_tran_flow.e12.attrs[
            "standard_name"
        ] = "horizontal scale factor along the transect at the normal velocity point"

    def plot_normal_velocity(self, time, plot_info: dict, cmap, smoothing_window=0):
        """
        Quick plot routine of velocity across the transect AB at a specific time.
        An option is provided to smooth the velocities along the transect.
        NOTE: For smoothing use even integers to smooth the x and y velocities together



        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value, 'vmin':value, 'vmax':value}
        Note that if vmin and max are not set then the colourbar will be centred at zero
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing, recommended
        to use even integers
        # TODO Add cmap definition to docstring.


        """
        debug(f"Plotting normal velocity for {get_slug(self)} with plot_info {plot_info}")
        try:
            data = self.data_cross_tran_flow.sel(t_dim=time)
        except KeyError:
            data = self.data_cross_tran_flow.isel(t_dim=time)

        if smoothing_window != 0:
            normal_velocities, depth = Transect.interpolate_slice(data.normal_velocities, data.depth_0)
            normal_velocities = Transect.moving_average(normal_velocities, smoothing_window, axis=-1)
            r_dim_2d = np.broadcast_to(data.r_dim, normal_velocities.shape)
        else:
            normal_velocities = data.normal_velocities
            depth = data.depth_0
            _, r_dim_2d = xr.broadcast(depth, data.r_dim)

        import matplotlib.pyplot as plt

        plt.close("all")
        fig = plt.figure(figsize=plot_info["fig_size"])
        ax = fig.gca()

        plt.pcolormesh(r_dim_2d, depth, normal_velocities, cmap=cmap)

        plt.title(plot_info["title"])
        plt.ylabel("Depth [m]")
        try:
            plt.clim(vmin=plot_info["vmin"], vmax=plot_info["vmax"])
        except KeyError:
            lim = np.nanmax(np.abs(normal_velocities))
            plt.clim(vmin=-lim, vmax=lim)
        plt.xticks([0, data.r_dim.values[-1]], ["A", "B"])
        plt.colorbar(label="Velocities across AB [m/s]")
        plt.gca().invert_yaxis()

        plt.show()
        return fig, ax

    def plot_depth_integrated_transport(self, time, plot_info: dict, smoothing_window=0):
        """
            Quick plot routine of depth integrated transport across the transect AB at a specific time.
            An option is provided to smooth along the transect via convolution,
            NOTE: For smoothing use even integers to smooth the x and y velocities together

        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value}
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing.
        Recommended to use even integers.

        Returns:
            pyplot object
        """
        debug(f"Generating quick plot for {get_slug(self)} with plot_info {plot_info}")
        try:
            data = self.data_cross_tran_flow.sel(t_dim=time)
        except KeyError:
            data = self.data_cross_tran_flow.isel(t_dim=time)

        if smoothing_window != 0:
            transport = self.moving_average(data.normal_transports, smoothing_window, axis=-1)
        else:
            transport = data.normal_transports

        import matplotlib.pyplot as plt

        plt.close("all")
        fig = plt.figure(figsize=plot_info["fig_size"])
        ax = fig.gca()

        plt.plot(data.r_dim, transport)

        plt.title(plot_info["title"])
        plt.xticks([0, data.r_dim[-1]], ["A", "B"])
        plt.ylabel("Volume transport across AB [SV]")
        plt.show()
        return fig, ax


class TransectT(Transect):
    """
    Class defining a transect on the t-grid, which is a 3d dataset along
    a linear path between a point A and a point B, with a time dimension,
    a depth dimension and an along transect dimension. The model Data on t-grid
    is subsetted in its entirety along these dimensions.

    Note that Point A should be closer to the southern boundary of the model domain.

    The user can either supply the start and end (lat,lon) coordinates of the
    transect, point_A and point_B respectively, or the model y, x indices defining it.
    In the latter case the user must ensure that the indices define a continuous
    transect, e.g. y=[10,11,11,12], x=[5,5,6,6].
    Only limited checks are performed on the suitability of the indices.

    Example usage:
        point_A = (54,-15)
        point_B = (56,-12)
        transect = coast.Transect_t( gridded_t, point_A, point_B )
        or
        transect = coast.Transect_t( gridded_t, y_indices=y_ind, x_indices=x_ind )

    Parameters
    ----------
    gridded_t : GRIDDED object on the t-grid
    point_a : tuple, (lat,lon)
    point_b : tuple, (lat,lon)
    y_indices : 1d array of model y indices defining the points of the transect
    x_indices : 1d array of model x indices defining the points of the transect

    """

    def __init__(self, gridded_t: Coast, point_a: tuple = None, point_b: tuple = None, y_indices=None, x_indices=None):
        super().__init__(gridded_t, point_a, point_b, y_indices, x_indices)

    def construct_pressure(self, ref_density=None, z_levels=None, extrapolate=False):
        """
            This method is for calculating the hydrostatic and surface pressure fields
            on horizontal levels in the vertical (z-levels). The motivation
            is to enable the calculation of horizontal gradients; however,
            the variables can quite easily be interpolated onto the original
            vertical grid.

            Requirements: The object's t-grid dataset must contain the sea surface height,
            Practical Salinity and the Potential Temperature variables.
            The GSW package is used to calculate the Absolute Pressure,
            Absolute Salinity and Conservate Temperature.

            Three new variables (density, hydrostatic pressure, surface pressure)
            are created and added to the Transect_t.data dataset:
                density_zlevels       (t_dim, depth_z_levels, r_dim)
                pressure_h_zlevels    (t_dim, depth_z_levels, r_dim)
                pressure_s            (t_dim, r_dim)

            Note that density is constructed using the EOS10
            equation of state.

        Parameters
        ----------
        ref_density: float
            reference density value, if None, then the transect mean across time,
            depth and along transect will be used.
        z_levels : (optional) numpy array
            1d array that defines the depths to interpolate the density and pressure
            on to. If not supplied, the Transect.gen_z_levels method will be used.
        extrapolate : boolean, default False
            If true the variables are extrapolated to the deepest level, if false,
            values below the bathymetry are set to NaN


        """

        # If there is no time dimension, add one, this is so
        # indexing can assume a time dimension exists
        if "t_dim" not in self.data.dims:
            self.data = self.data.expand_dims(dim={"t_dim": 1}, axis=0)

        # Generate vertical levels if not supplied
        if z_levels is None:
            z_levels = Transect.gen_z_levels(self.data.bathymetry.max().item())

        shape_ds = (self.data.t_dim.size, len(z_levels), self.data.r_dim.size)
        salinity_z = np.ma.zeros(shape_ds)
        temperature_z = np.ma.zeros(shape_ds)
        salinity_s = self.data.salinity.to_masked_array()
        temperature_s = self.data.temperature.to_masked_array()
        s_levels = self.data.depth_0.values

        # Interpolate salinity and temperature onto z-levels
        # Note. At the current time there does not appear to be a good algorithm for
        # performing this type of interpolation without loops, which can be a bottleneck.
        # Griddata is an option but does not support extrapolation and did not
        # have noticable performance benefit.
        for it in self.data.t_dim:
            for ir in self.data.r_dim:
                if not np.all(np.isnan(salinity_s[it, :, ir].data)):
                    # Need to remove the levels below the (envelope) bathymetry which are NaN
                    salinity_s_r = salinity_s[it, :, ir].compressed()
                    temperature_s_r = temperature_s[it, :, ir].compressed()
                    s_levels_r = s_levels[: len(salinity_s_r), ir]

                    sal_func = interpolate.interp1d(s_levels_r, salinity_s_r, kind="linear", fill_value="extrapolate")
                    temp_func = interpolate.interp1d(
                        s_levels_r, temperature_s_r, kind="linear", fill_value="extrapolate"
                    )

                    if extrapolate is True:
                        salinity_z[it, :, ir] = sal_func(z_levels)
                        temperature_z[it, :, ir] = temp_func(z_levels)
                    else:
                        # set levels below the bathymetry to nan
                        salinity_z[it, :, ir] = np.where(
                            z_levels <= self.data.bathymetry.values[ir], sal_func(z_levels), np.nan
                        )
                        temperature_z[it, :, ir] = np.where(
                            z_levels <= self.data.bathymetry.values[ir], temp_func(z_levels), np.nan
                        )

        if extrapolate is False:
            # remove redundent levels
            active_z_levels = np.count_nonzero(~np.isnan(salinity_z), axis=1).max()
            salinity_z = salinity_z[:, :active_z_levels, :]
            temperature_z = temperature_z[:, :active_z_levels, :]
            z_levels = z_levels[:active_z_levels]

        # Absolute Pressure (depth must be negative)
        pressure_absolute = np.ma.masked_invalid(gsw.p_from_z(-z_levels[:, np.newaxis], self.data.latitude.values))
        # Absolute Salinity
        salinity_absolute = np.ma.masked_invalid(
            gsw.SA_from_SP(salinity_z, pressure_absolute, self.data.longitude.values, self.data.latitude.values)
        )
        salinity_absolute = np.ma.masked_less(salinity_absolute, 0)
        # Conservative Temperature
        temp_conservative = np.ma.masked_invalid(gsw.CT_from_pt(salinity_absolute, temperature_z))
        # In-situ density
        density_z = np.ma.masked_invalid(gsw.rho(salinity_absolute, temp_conservative, pressure_absolute))

        coords = {
            "depth_z_levels": ("depth_z_levels", z_levels),
            "latitude": ("r_dim", self.data.latitude.values),
            "longitude": ("r_dim", self.data.longitude.values),
        }
        dims = ["depth_z_levels", "r_dim"]
        attributes = {"units": "kg / m^3", "standard name": "In-situ density on the z-level vertical grid"}

        if shape_ds[0] != 1:
            coords["time"] = ("t_dim", self.data.time.values)
            dims.insert(0, "t_dim")

        if ref_density is None:
            ref_density = np.mean(density_z)
        self.data["density_zlevels"] = xr.DataArray(np.squeeze(density_z), coords=coords, dims=dims, attrs=attributes)

        # Cumulative integral of perturbation density on z levels
        density_cumulative = -cumtrapz(density_z - ref_density, x=-z_levels, axis=1, initial=0)
        hydrostatic_pressure = density_cumulative * self.gravity

        attributes = {
            "units": "kg m^{-1} s^{-2}",
            "standard name": "Hydrostatic perturbation pressure on the z-level vertical grid",
        }
        self.data["pressure_h_zlevels"] = xr.DataArray(
            np.squeeze(hydrostatic_pressure), coords=coords, dims=dims, attrs=attributes
        )
        self.data["pressure_s"] = ref_density * self.gravity * self.data.ssh.squeeze()
        self.data.pressure_s.attrs = {"units": "kg m^{-1} s^{-2}", "standard_name": "Surface perturbation pressure"}



"""
Python definitions used to help with plotting routines.

*Methods Overview*
    -> geo_scatter(): Geographical scatter plot.
"""

import matplotlib.pyplot as plt
from warnings import warn
from .logging_util import warn
import numpy as np


def r2_lin(x, y, fit):
    """For calculating r-squared of a linear fit. Fit should be a python polyfit object."""
    y_estimate = fit(x)
    difference = (y - y_estimate) ** 2
    y_mean = np.nanmean(y)
    mean_square_deviation = (y - y_mean) ** 2

    total_deviation = np.nansum(mean_square_deviation)
    residual = np.nansum(difference)

    correlation_coefficient = 1 - residual / total_deviation

    return correlation_coefficient


def scatter_with_fit(x, y, s=10, c="k", yex=True, dofit=True):
    """Does a scatter plot with a linear fit. Will also draw y=x for
    comparison.

    Parameters
    ----------
    x     : (array) Values for the x-axis
    y     : (array) Values for the y-axis
    s     : (float or array) Marker size(s)
    c     : (float or array) Marker colour(s)
    yex   : (bool) True to plot y=x
    dofit : (bool) True to calculate and plot linear fit

    Returns
    -------
    Figure and axis objects for further customisation

    Example Useage
    -------
    x = np.arange(0,50)
    y = np.arange(0,50)/1.5
    f,a = scatter_with_fit(x,y)
    a.set_title('Example scatter with fit')
    a.set_xlabel('Example x axis')
    a.set_ylabel('Example y axis')
    """

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    x = np.ma.masked_invalid(x)
    y = np.ma.masked_invalid(y)
    combined_mask = np.ma.mask_or(x.mask, y.mask)
    x.mask = combined_mask
    y.mask = combined_mask

    xmax = np.ma.max(x)
    xmin = np.ma.min(x)
    ymax = np.ma.max(y)
    ymin = np.ma.min(y)
    axmax0 = np.max([xmax, ymax])
    axmin0 = np.min([xmin, ymin])
    axmin = axmin0 - 0.1 * np.abs(axmax0 - axmin0)
    axmax = axmax0 + 0.1 * np.abs(axmax0 - axmin0)

    if yex:
        line_x = [axmin, axmax]
        fit_yx = np.poly1d([1, 0])
        ax.plot(line_x, fit_yx(line_x), c=[0.5, 0.5, 0.5], linewidth=1)

    ax.scatter(x, y, c=c, s=s)

    if dofit:
        line_x = [axmin, axmax]
        # Calculate data fit and cast to poly1d object
        fit_tmp = np.ma.polyfit(x, y, 1)
        fit = np.poly1d(fit_tmp)
        ax.plot(line_x, fit(line_x), c=[1, 128 / 255, 0], linewidth=1.5)
        r2 = r2_lin(x, y, fit)

    ax.set_xlim(axmin, axmax)
    ax.set_ylim(axmin, axmax)
    ax.set_aspect("equal", adjustable="box")
    ax.grid()

    if dofit:
        ax.text(
            0.4, 0.125, "{} {:03.2f} {} {:03.2f}".format("y =", fit_tmp[0], "x +", fit_tmp[1]), transform=ax.transAxes
        )
        ax.text(0.4, 0.05, "{} {:03.2f} ".format("$R^2$ =", r2), transform=ax.transAxes)

    return fig, ax


def create_geo_subplots(lonbounds, latbounds, n_r=1, n_c=1, figsize=(7, 7)):
    """
    A routine for creating an axis for any geographical plot. Within the
    specified longitude and latitude bounds, a map will be drawn up using
    cartopy. Any type of matplotlib plot can then be added to this figure.
    For example:
    Example Useage
    #############
        f,a = create_geo_axes(lonbounds, latbounds)
        sca = a.scatter(stats.longitude, stats.latitude, c=stats.corr,
                        vmin=.75, vmax=1,
                        edgecolors='k', linewidths=.5, zorder=100)
        f.colorbar(sca)
        a.set_title('SSH correlations \n Monthly PSMSL tide gauge vs CO9_AMM15p0',
                    fontsize=9)
    * Note: For scatter plots, it is useful to set zorder = 100 (or similar
            positive number)
    """

    import cartopy.crs as ccrs  # mapping plots
    from cartopy.feature import NaturalEarthFeature

    # If no figure or ax is provided, create a new one
    # fig = plt.figure()
    # fig.clf()
    fig, ax = plt.subplots(
        n_r, n_c, subplot_kw={"projection": ccrs.PlateCarree()}, sharey=True, sharex=True, figsize=figsize
    )
    land_color = [0.9, 0.9, 0.9]
    coast_color = [0, 0, 0]
    coast_width = 0.25

    if n_r * n_c > 1:
        ax = ax.flatten()
        for rr in range(n_r * n_c):
            coast = NaturalEarthFeature(category="physical", facecolor=land_color, name="coastline", scale="50m")
            ax[rr].add_feature(coast, edgecolor=coast_color, linewidth=coast_width)
            # ax.coastlines(facecolor=[0.8,0.8,0.8])
            gl = ax[rr].gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linewidth=0.5, color="gray", linestyle="-")
            gl.top_labels = False
            gl.right_labels = False
            if rr % n_c == 0:
                gl.left_labels = True
            else:
                gl.left_labels = False

            if np.abs(n_r * n_c - rr) <= n_c:
                gl.bottom_labels = True
            else:
                gl.bottom_labels = False
            ax[rr].set_xlim(lonbounds[0], lonbounds[1])
            ax[rr].set_ylim(latbounds[0], latbounds[1])
            ax[rr].set_aspect("auto")

        ax = ax.reshape((n_r, n_c))
    else:
        coast = NaturalEarthFeature(category="physical", facecolor=land_color, name="coastline", scale="50m")
        ax.add_feature(coast, edgecolor=coast_color, linewidth=coast_width)
        # ax.coastlines(facecolor=[0.8,0.8,0.8])
        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linewidth=0.5, color="gray", linestyle="-")
        gl.top_labels = False
        gl.right_labels = False

        gl.left_labels = True
        gl.bottom_labels = True

        ax.set_xlim(lonbounds[0], lonbounds[1])
        ax.set_ylim(latbounds[0], latbounds[1])
        ax.set_aspect("auto")

    plt.show()
    return fig, ax


def create_geo_axes(lonbounds, latbounds):
    """
    A routine for creating an axis for any geographical plot. Within the
    specified longitude and latitude bounds, a map will be drawn up using
    cartopy. Any type of matplotlib plot can then be added to this figure.
    For example:

    Example Useage
    #############

        f,a = create_geo_axes(lonbounds, latbounds)
        sca = a.scatter(stats.longitude, stats.latitude, c=stats.corr,
                        vmin=.75, vmax=1,
                        edgecolors='k', linewidths=.5, zorder=100)
        f.colorbar(sca)
        a.set_title('SSH correlations \n Monthly PSMSL tide gauge vs CO9_AMM15p0',
                    fontsize=9)

    * Note: For scatter plots, it is useful to set zorder = 100 (or similar
            positive number)
    """

    import cartopy.crs as ccrs  # mapping plots
    from cartopy.feature import NaturalEarthFeature

    # If no figure or ax is provided, create a new one
    fig = plt.figure(1)
    fig.clf()
    ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())

    coast = NaturalEarthFeature(category="physical", facecolor=[0.9, 0.9, 0.9], name="coastline", scale="50m")
    ax.add_feature(coast, edgecolor="gray")
    # ax.coastlines(facecolor=[0.8,0.8,0.8])
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linewidth=0.5, color="gray", linestyle="-")
    gl.top_labels = False
    gl.bottom_labels = True
    gl.right_labels = False
    gl.left_labels = True

    ax.set_xlim(lonbounds[0], lonbounds[1])
    ax.set_ylim(latbounds[0], latbounds[1])
    ax.set_aspect("auto")

    plt.show()
    return fig, ax


def ts_diagram(temperature, salinity, depth):

    fig = plt.figure(figsize=(10, 7))
    ax = plt.scatter(salinity, temperature, c=depth)
    cbar = plt.colorbar()
    cbar.set_label("Depth (m)")
    plt.title("T-S Diagram")
    plt.xlabel("Salinity")
    plt.ylabel("Temperature")

    return fig, ax


def geo_scatter(
    longitude,
    latitude,
    c=None,
    s=None,
    scatter_kwargs=None,
    coastline_kwargs=None,
    gridline_kwargs=None,
    figure_kwargs={},
    title="",
    figsize=None,
):  # TODO Some unused parameters here
    """
    Uses CartoPy to create a geographical scatter plot with land boundaries.

        Parameters
        ----------
        longitude : (array) Array of longitudes of marker locations
        latitude  : (array) Array of latitudes of marker locations
        colors    : (array) Array of values to use for colouring markers
        title     : (str) Plot title, to appear at top of figure
        xlim      : (tuple) Tuple of limits to apply to the x-axis (longitude axis)
        ylim      : (tuple) Limits to apply to the y-axis (latitude axis)

        Returns
        -------
        Figure and axis objects for further customisation

    """
    try:
        import cartopy.crs as ccrs  # mapping plots
        from cartopy.feature import NaturalEarthFeature
    except ImportError:
        import sys

        warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
        sys.exit(-1)

    if coastline_kwargs is None:
        coastline_kwargs = {"facecolor": [0.9, 0.9, 0.9], "name": "coastline", "scale": "50m"}
    if scatter_kwargs is None:
        scatter_kwargs = {}
    fig = plt.figure(**figure_kwargs)

    ax = plt.subplot(111, projection=ccrs.PlateCarree())
    sca = ax.scatter(longitude, y=latitude, c=c, s=s, zorder=100, **scatter_kwargs)
    coast = NaturalEarthFeature(category="physical", **coastline_kwargs)
    ax.add_feature(coast, edgecolor="gray")
    # ax.coastlines(facecolor=[0.8,0.8,0.8])
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linewidth=0.5, color="gray", linestyle="-")
    gl.top_labels = False
    gl.bottom_labels = True
    gl.right_labels = False
    gl.left_labels = True
    plt.title(title)

    if c is not None and "vmax" in scatter_kwargs.keys() and "vmin" in scatter_kwargs.keys():
        extend_max = np.nanmax(c) > scatter_kwargs["vmax"]
        extend_min = np.nanmin(c) < scatter_kwargs["vmin"]
        extend = "neither"
        if extend_max and extend_min:
            extend = "both"
        if extend_max and not extend_min:
            extend = "max"
        if not extend_max and extend_min:
            extend = "min"
    else:
        extend = "neither"

    plt.colorbar(sca, extend=extend)
    ax.set_aspect("auto")

    plt.show()
    return fig, ax


def determine_colorbar_extension(color_data, vmin, vmax):
    """Can be used to automatically determine settings for colorbar
    extension arrows. Color_data is the data used for the colormap, vmin
    and vmax are the colorbar limits. Will output a string: "both", "max",
    "min" or "neither", which can be inserted straight into a call to
    matplotlib.pyplot.colorbar().
    """
    extend_max = np.nanmax(color_data) > vmax
    extend_min = np.nanmin(color_data) < vmin

    if extend_max and extend_min:
        return "both"
    elif extend_max and not extend_min:
        return "max"
    elif not extend_max and extend_min:
        return "min"
    else:
        return "neither"


def determine_clim_by_standard_deviation(color_data, n_std_dev=2.5):
    """Automatically determine color limits based on number of standard
    deviations from the mean of the color data (color_data). Useful if there
    are outliers in the data causing difficulties in distinguishing most of
    the data. Outputs vmin and vmax which can be passed to plotting routine
    or plt.clim().
    """
    color_data_mean = np.nanmean(color_data)
    color_data_std = np.nanstd(color_data)
    vmin = color_data_mean - n_std_dev * color_data_std
    vmax = color_data_mean + n_std_dev * color_data_std
    return vmin, vmax



"""
Python definitions used to aid in the calculation of Continuous Ranked
Probability Score.
*Methods Overview*
    -> crps_sonf_fixed(): Single obs neighbourhood forecast CRPS for fixed obs
    -> crps_song_moving(): Same as above for moving obs
"""

import numpy as np
from . import general_utils


def crps_empirical(sample, obs):
    """Calculates CRPS for a single observations against a sample of values.
    This sample of values may be an ensemble of model forecasts or a model
    neighbourhood. This is a comparison of a Heaviside function defined by
    the observation value and an Empirical Distribution Function (EDF)
    defined by the sample of values. This sample is sorted to create the
    EDF.

    The calculation method is that outlined by Hersbach et al. (2000).
    Each member of a supplied sample is weighted equally.

    Args:
        sample (array): Array of points (ensemble or neighbourhood)
        xa (float): A single 'observation' value which to compare against
                    sample CDF.
    Returns:
        A single CRPS value.
    """

    def calc(alpha, beta, p):  # TODO It would be better to define this outside of the function
        return alpha * p**2 + beta * (1 - p) ** 2  # TODO Could this be a lambda?

    xa = float(obs)
    crps_integral = 0
    sample = np.array(sample)

    if all(np.isnan(sample)) or np.isnan(obs):
        return np.nan

    sample = sample[~np.isnan(sample)]
    sample = np.sort(sample)
    sample_size = len(sample)

    alpha = np.zeros(sample_size - 1)
    beta = np.zeros(sample_size - 1)
    # sample[1:] = upper bounds, and vice versa

    tmp = sample[1:] - sample[:-1]
    tmp_logic = sample[1:] < xa
    alpha[tmp_logic] = tmp[tmp_logic]

    tmp_logic = sample[:-1] > xa
    beta[tmp_logic] = tmp[tmp_logic]

    tmp_logic = (sample[1:] > xa) * (sample[:-1] < xa)
    tmp = xa - sample[:-1]
    alpha[tmp_logic] = tmp[tmp_logic]
    tmp = sample[1:] - xa
    beta[tmp_logic] = tmp[tmp_logic]

    p = np.arange(1, sample_size) / sample_size
    c = alpha * p**2 + beta * (1 - p) ** 2
    crps_integral = np.sum(c)

    # Intervals 0 and N, where p = 0 and 1 respectively
    if xa < sample[0]:
        p = 0
        alpha = 0
        beta = sample[0] - xa
        crps = calc(alpha, beta, p)
        crps_integral += crps
    elif xa > sample[-1]:
        p = 1
        alpha = xa - sample[-1]
        beta = 0
        crps = calc(alpha, beta, p)
        crps_integral += crps

    crps = crps_integral

    return crps


def crps_empirical_loop(sample, obs):
    """Like crps_empirical, however a loop is used instead of numpy
    boolean indexing. For large samples, will be slower but consume less
    memory.
    """

    def calc(alpha, beta, p):  # TODO It would be better to define this outside of the function
        return alpha * p**2 + beta * (1 - p) ** 2  # TODO Could this be a lambda?

    crps_integral = 0
    sample = np.array(sample)
    sample = sample[~np.isnan(sample)]
    sample = np.sort(sample)
    sample_size = len(sample)

    # All intervals within range of the sample distribution
    for ii in range(0, sample_size - 1):
        p = (ii + 1) / sample_size
        if obs > sample[ii + 1]:
            alpha = sample[ii + 1] - sample[ii]
            beta = 0
        elif obs < sample[ii]:
            alpha = 0
            beta = sample[ii + 1] - sample[ii]
        else:
            alpha = obs - sample[ii]
            beta = sample[ii + 1] - obs
        crps = calc(alpha, beta, p)
        crps_integral += crps
    # Intervals 0 and N, where p = 0 and 1 respectively
    if obs < sample[0]:
        p = 0
        alpha = 0
        beta = sample[0] - obs
        crps = calc(alpha, beta, p)
        crps_integral += crps
    elif obs > sample[-1]:
        p = 1
        alpha = obs - sample[-1]
        beta = 0
        crps = calc(alpha, beta, p)
        crps_integral += crps

    return crps_integral


def crps_sonf_fixed(
    mod_array,
    obs_lon,
    obs_lat,
    obs_var,
    obs_time,
    nh_radius: float,
    time_interp: str,
):
    """
    Handles the calculation of single-observation neighbourhood forecast CRPS
    for a time series at a fixed observation location. Differs from
    crps_sonf_moving in that it only need calculate a model neighbourhood once.
    Parameters
    ----------
    mod_array   : (xarray DataArray) DataArray from a Model Dataset
    obs_lon     : (float) Longitude of fixed observation point
    obs_lat     : (float) Latitude of fixed observation point
    obs_var     : (array) of floatArray of variable values, e.g time series
    obs_time    : (array) of datetimeArray of times, corresponding to obs_var
    nh_radius   : (float) Neighbourhood radius in km
    time_interp : (str) Type of time interpolation to use
    Returns
    -------
    crps_list     : Array of CRPS values
    n_model_pts   : Array containing the number of model points used for
                    each CRPS value
    contains_land : Array of bools indicating where a model neighbourhood
                    contained land.
    """

    # Define output arrays
    n_neighbourhoods = obs_var.shape[0]
    crps_list = np.zeros(n_neighbourhoods) * np.nan
    n_model_pts = np.zeros(n_neighbourhoods) * np.nan
    contains_land = np.zeros(n_neighbourhoods, dtype=bool)

    # Loop over neighbourhoods
    neighbourhood_indices = np.arange(0, n_neighbourhoods)

    # Get model neighbourhood subset using specified method
    subset_ind = general_utils.subset_indices_by_distance(
        mod_array.longitude.values, mod_array.latitude.values, obs_lon, obs_lat, nh_radius
    )
    mod_subset = mod_array.isel(y_dim=subset_ind[0], x_dim=subset_ind[1])
    mod_subset = mod_subset.swap_dims({"t_dim": "time"})

    # Check that the model neighbourhood contains points
    if subset_ind[0].shape[0] == 0 or subset_ind[1].shape[0] == 0:
        crps_list = crps_list * np.nan
    else:
        # Subset model data in time and space: model -> obs
        for ii in neighbourhood_indices:

            mod_subset_time = mod_subset.interp(
                time=obs_time[ii], method=time_interp, kwargs={"fill_value": "extrapolate"}
            )

            # Check if neighbourhood contains a land value (TODO:mask)
            if any(np.isnan(mod_subset_time)):
                contains_land[ii] = True
            # Check that neighbourhood contains a value
            if all(np.isnan(mod_subset_time)):
                pass
            else:
                # Calculate CRPS and put into output array
                crps_list[ii] = crps_empirical(mod_subset_time.values, obs_var[ii])
                n_model_pts[ii] = int(mod_subset.shape[0])

    return crps_list, n_model_pts, contains_land


def crps_sonf_moving(mod_array, obs_lon, obs_lat, obs_var, obs_time, nh_radius: float, time_interp: str, obs_batch=10):
    """
    Handles the calculation of single-observation neighbourhood forecast CRPS
    for a moving observation instrument. Differs from crps_sonf_fixed in that
    latitude and longitude are arrays of locations. Mod_array must contain
    dimensions x_dim, y_dim and t_dim and coordinates longitude, latitude,
    time.
    Parameters
    ----------
    mod_array   : (xarray DataArray) DataArray from a Model Dataset
    obs_lon     : (1Darray) Longitudes of fixed observation point
    obs_lat     : (1Darray) Latitudes of fixed observation point
    obs_var     : (1Darray) of floatArray of variable values, e.g time series
    obs_time    : (1Darray) of datetimeArray of times, corresponding to obs_var
    nh_radius   : (float) Neighbourhood radius in km
    time_interp : (str) Type of time interpolation to use
    Returns
    -------
    crps_list     : Array of CRPS values
    n_model_pts   : Array containing the number of model points used for
                    each CRPS value
    contains_land : Array of bools indicating where a model neighbourhood
                    contained land.
    """

    # Define output arrays
    n_neighbourhoods = obs_var.shape[0]
    crps_list = np.zeros(n_neighbourhoods) * np.nan
    n_model_pts = np.zeros(n_neighbourhoods) * np.nan
    contains_land = np.zeros(n_neighbourhoods, dtype=bool)
    # Loop over neighbourhoods
    neighbourhood_indices = np.arange(0, n_neighbourhoods)
    for ii in neighbourhood_indices:
        # Neighbourhood centre
        cntr_lon = obs_lon[ii]
        cntr_lat = obs_lat[ii]

        # Get model neighbourhood subset using specified method
        subset_ind = general_utils.subset_indices_by_distance(
            mod_array.longitude, mod_array.latitude, cntr_lon, cntr_lat, nh_radius
        )
        # Check that the model neighbourhood contains points
        if subset_ind[0].shape[0] == 0 or subset_ind[1].shape[0] == 0:
            crps_list[ii] = np.nan
        else:
            # Subset model data in time and space: model -> obs
            mod_subset = mod_array.isel(y_dim=subset_ind[0], x_dim=subset_ind[1])
            mod_subset = mod_subset.swap_dims({"t_dim": "time"})
            mod_subset = mod_subset.interp(time=obs_time[ii], method=time_interp, kwargs={"fill_value": "extrapolate"})

            # Check if neighbourhood contains a land value (TODO:mask)
            if any(np.isnan(mod_subset)):
                contains_land[ii] = True
            # Check that neighbourhood contains a value
            if all(np.isnan(mod_subset)):
                pass
            else:
                # Calculate CRPS and put into output array
                crps_list[ii] = crps_empirical(mod_subset, obs_var[ii])
                n_model_pts[ii] = int(mod_subset.shape[0])

    return crps_list, n_model_pts, contains_land



"""Timeseries Class"""
from .index import Indexed
from . import general_utils


class Timeseries(Indexed):
    """Parent class for Tidegauge and other timeseries type datasets
    Common methods ...
    """

    pass



from setuptools import setup
from sys import argv
from types import SimpleNamespace


PACKAGE = SimpleNamespace(
    **{
        "name": "COAsT",
        "version": "2.0.2",
        "description": "This is the Coast Ocean Assessment Tool",
        "url": "https://www.bodc.ac.uk",
        "download_url": "https://github.com/British-Oceanographic-Data-Centre/COAsT/",
        "author": "British Oceanographic Data Centre (BODC)",
        "author_email": "bodcsoft@bodc.ac.uk",
        "license": "Put something here",
        "license_family": "OTHER",
        "setup_requires": ["wheel"],
        "classifiers": [
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Hydrology",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
        ],
        "keywords": ["NEMO", "shallow water", "ocean assessment"],
        "project_urls": {"documentation": "https://british-oceanographic-data-centre.github.io/COAsT/"},
        "install_requires": [
            "numpy>=1.22.3",
            "dask>=2022.3.0",
            "dask[complete]>=2022.3.0",
            "xarray>=2022.3.0",
            "matplotlib>=3.5.1",
            "netCDF4>=1.5.8",
            "scipy>=1.8.0",
            "gsw>=3.4.0",
            "utide>=0.2.6",
            "scikit-learn>=1.0.2",
            "scikit-image>=0.19.2",
            "statsmodels>=0.13.2",
        ],
        "python_requires": ">=3.8",
        "packages": ["coast"],
        "include_package_data": True,
        "github": "British-Oceanographic-Data-Centre",
    }
)


def generate_conda(directory="conda"):
    import oyaml as yaml  # OYaml is used to preserve the order of the metadata dict in YAML output
    from collections import OrderedDict
    from os import path

    requirements = PACKAGE.install_requires + ["python", "pip"]

    package_metadata = OrderedDict(
        {
            "package": {"name": PACKAGE.name.lower(), "version": PACKAGE.version},
            "source": {
                "url": f"https://pypi.io/packages/source/{PACKAGE.name[0]}/{PACKAGE.name}/{PACKAGE.name}-{PACKAGE.version}.tar.gz"
            },
            "build": {"number": 0, "script": "{{ PYTHON }} -m pip install . --no-deps --ignore-installed -vv "},
            "requirements": {"host": requirements, "run": requirements},
            "test": {"imports": ["coast"]},
            "about": {
                "home": PACKAGE.url,
                "license": PACKAGE.license,
                "license_family": PACKAGE.license_family,
                "license_file": "",
                "summary": PACKAGE.description,
                "doc_url": "",
                "dev_url": "",
            },
            "extra": {"recipe-maintainers": [PACKAGE.github]},
        }
    )

    yaml_path = path.join(directory, "meta.yaml")
    yaml.Dumper.ignore_aliases = lambda *args: True  # Dummy function to return true
    with open(yaml_path, "w") as meta_file:
        yaml.dump(package_metadata, meta_file)


if __name__ == "__main__":
    if argv[1] == "conda":
        generate_conda()
    else:
        setup(
            name=PACKAGE.name,
            version=PACKAGE.version,
            description=PACKAGE.description,
            url=PACKAGE.url,
            download_url=PACKAGE.download_url,
            author=PACKAGE.author,
            author_email=PACKAGE.author_email,
            license=PACKAGE.license,  # TODO,
            setup_requires=PACKAGE.setup_requires,
            classifiers=PACKAGE.classifiers,
            keywords=PACKAGE.keywords,
            project_urls=PACKAGE.project_urls,
            install_requires=PACKAGE.install_requires,
            python_requires=PACKAGE.python_requires,
            packages=PACKAGE.packages,
            include_package_data=PACKAGE.include_package_data,
        )



import xarray as xr
import numpy as np
from scipy import linalg
from scipy.signal import hilbert


def compute_eofs(variable: xr.DataArray, full_matrices: bool = False, time_dim_name: str = "t_dim"):
    """
    Computes the Empirical Orthogonal Functions (EOFs) of a variable (time series)
    that has 3 dimensions where one is time, i.e. (x,y,time)

    Returns the set of EOF modes, the associated temporal projections and the
    variance explained by each mode as DataArrays within an xarray Dataset.

    All-NaN time series, such as those at land points, are handled and ignored;
    however, isolated NaNs within a time series, i.e. missing data point, must
    be filled before calling the function.

    The variable will be de-meaned in time before the EOFs are computed, normalisation
    should be carried out before calling the function if desired. The returned EOFs and
    temporal projections are not scaled or normalised.

    Parameters
    ----------
    variable : (xarray.DataArray), 3-dimensional variable of size (I,J,T),
    containing I*J time series
    full_matrices : (boolean, default False) if false computes only first K EOFs
    where K=min(I*J,T), where T is total number of time points. Setting to True
    could demand significant memory.
    time_dim_name : (string, default 't_dim') the name of the time dimension.

    Returns
    -------
    dataset : xarray Dataset, containing the EOFs, temporal projections and
    variance explained as xarray DataArrays. The relevent coordinates
    from the original data variable are also included

    """

    variable = variable.transpose(..., time_dim_name, transpose_coords=False)
    i, j, t = np.shape(variable.data)  # TODO Can someone sciencey give me the proper name for these?
    signal = np.reshape(variable.data, (i * j, t))

    # Remove constant zero point such as land points
    active_ind = np.where(np.nan_to_num(signal).any(axis=1))[0]
    signal = signal[active_ind, :]
    # Remove time mean at each grid point
    mean = signal.mean(axis=1)
    signal = signal - mean[:, np.newaxis]

    eofs, projections, variance_explained, mode_count = _compute(signal, full_matrices, active_ind, i * j)

    eofs = np.reshape(eofs, (i, j, mode_count))

    # Assign to xarray variables
    # copy the coordinates
    coords = {"mode": ("mode", np.arange(1, mode_count + 1))}
    time_coords = {"mode": ("mode", np.arange(1, mode_count + 1))}
    for coord in variable.coords:
        if variable.dims[2] not in variable[coord].dims:
            coords[coord] = (variable[coord].dims, variable[coord].values)
        else:
            if (variable.dims[2],) == variable[coord].dims:
                time_coords[coord] = (variable[coord].dims, variable[coord].values)

    dataset = xr.Dataset()

    dims = (variable.dims[:2]) + ("mode",)
    dataset["EOF"] = xr.DataArray(eofs, coords=coords, dims=dims)
    dataset.EOF.attrs["standard name"] = "EOF"

    dims = (variable.dims[2], "mode")
    dataset["temporal_proj"] = xr.DataArray(projections, coords=time_coords, dims=dims)
    dataset.temporal_proj.attrs["standard name"] = "temporal projection"

    dataset["variance"] = xr.DataArray(
        variance_explained, coords={"mode": ("mode", np.arange(1, mode_count + 1))}, dims=["mode"]
    )
    dataset.variance.attrs["standard name"] = "percentage of variance explained"

    return dataset


def compute_hilbert_eofs(variable: xr.DataArray, full_matrices=False, time_dim_name: str = "t_dim"):
    """
    Computes the complex Hilbert Empirical Orthogonal Functions (HEOFs) of a
    variable (time series) that has 3 dimensions where one is time, i.e. (x,y,time).
    See https://doi.org/10.1002/joc.1499

    Returns the set of HEOF amplitude and phase modes, the associated temporal
    projection amplitudes and phases and the variance explained by each mode
    as DataArrays within an xarray Dataset.

    All-NaN time series, such as those at land points, are handled and ignored;
    however, isolated NaNs within a time series, i.e. missing data point, must
    be filled before calling the function.

    The variable will be de-meaned in time before the EOFs are computed, normalisation
    should be carried out before calling the function if desired. The returned EOFs and
    temporal projections are not scaled or normalised.

    Parameters
    ----------
    variable : (xarray.DataArray), 3-dimensional variable of size (I,J,T),
    containing I*J time series
    full_matrices : (boolean, default False) if false computes only first K EOFs
    where K=min(I*J,T), where T is total number of time points.
    time_dim_name : (string, default 't_dim') the name of the time dimension.

    Returns
    -------
    dataset : xarray Dataset, containing the EOF amplitudes and phases,
    temporal projection amplitude and phases and the variance explained
    as xarray DataArrays. The relevent coordinates
    from the original data variable are also in the dataset.

    """
    variable = variable.transpose(..., time_dim_name, transpose_coords=False)
    i, j, t = np.shape(variable.data)  # TODO Can someone sciencey give me the proper name for these?
    signal = np.reshape(variable.data, (i * j, t))

    # Remove constant zero point such as land points
    active_ind = np.where(np.nan_to_num(signal).any(axis=1))[0]
    signal = signal[active_ind, :]
    # Remove time mean at each grid point
    mean = np.mean(signal, axis=1)
    signal = signal - mean[:, np.newaxis]
    # Apply Hilbert transform
    signal = hilbert(signal, axis=1)
    # Compute EOFs
    eofs, projections, variance_explained, mode_count = _compute(signal, full_matrices, active_ind, i * j)

    # Extract the amplitude and phase of the projections
    projection_amp = np.absolute(projections)
    projection_phase = np.angle(projections)

    # Extract the amplitude and phase of the spatial component
    eofs = np.reshape(eofs, (i, j, mode_count))
    eof_amp = np.absolute(eofs)
    eof_phase = np.angle(eofs)

    # Assign to xarray variables
    # copy the coordinates
    dataset = xr.Dataset()
    coords = {"mode": ("mode", np.arange(1, mode_count + 1))}
    time_coords = {"mode": ("mode", np.arange(1, mode_count + 1))}
    for coord in variable.coords:
        if variable.dims[2] not in variable[coord].dims:
            coords[coord] = (variable[coord].dims, variable[coord].values)
        else:
            if (variable.dims[2],) == variable[coord].dims:
                time_coords[coord] = (variable[coord].dims, variable[coord].values)

    dims = (variable.dims[:2]) + ("mode",)
    dataset["EOF_amp"] = xr.DataArray(eof_amp, coords=coords, dims=dims)
    dataset.EOF_amp.attrs["standard_name"] = "EOF amplitude"
    dataset["EOF_phase"] = xr.DataArray(np.rad2deg(eof_phase), coords=coords, dims=dims)
    dataset.EOF_phase.attrs["standard_name"] = "EOF phase"
    dataset.EOF_phase.attrs["units"] = "degrees"

    dims = (variable.dims[2], "mode")
    dataset["temporal_amp"] = xr.DataArray(projection_amp, coords=time_coords, dims=dims)
    dataset.temporal_amp.attrs["standard_name"] = "temporal projection amplitude"
    dataset["temporal_phase"] = xr.DataArray(np.rad2deg(projection_phase), coords=time_coords, dims=dims)
    dataset.temporal_phase.attrs["standard_name"] = "temporal projection phase"
    dataset.temporal_phase.attrs["units"] = "degrees"

    dataset["variance"] = xr.DataArray(
        variance_explained, coords={"mode": ("mode", np.arange(1, mode_count + 1))}, dims=["mode"]
    )
    dataset.variance.attrs["standard name"] = "percentage of variance explained"

    return dataset


def _compute(signal, full_matrices, active_ind, number_points):
    """
    Compute eofs, projections and variance explained using a Singular Value Decomposition

    Parameters
    ----------
    signal : (array) the signal
    full_matrices : (boolean) whether to return a full or abbreviated SVD
    active_ind : (array) indices of points with non-null signal
    number_points : (int) number of points in original data set

    Returns
    -------
    EOFs : (array) the EOFs in 2d form
    projections : (array) the projection of the EOFs
    variance_explained : (array) variance explained by each mode
    mode_count : (int) number of modes computed

    """
    p, d, q = linalg.svd(
        signal, full_matrices=full_matrices
    )  # TODO Can someone sciencey give me the proper name for these?
    mode_count = p.shape[-1]
    eofs = np.full((number_points, mode_count), np.nan, dtype=p.dtype)
    eofs[active_ind, :] = p

    # Calculate variance explained
    if full_matrices:
        variance_explained = 100.0 * (d**2 / np.sum(d**2))
    else:
        Inv = p.dot(np.dot(np.diag(d), q))  # PDQ
        var1 = np.sum(np.var(Inv, axis=1, ddof=1))
        var2 = np.sum(np.var(signal, axis=1, ddof=1))
        mult = var1 / var2
        variance_explained = 100.0 * mult * (d**2 / np.sum(d**2))

    # Extract EOF projections
    projections = np.transpose(q) * d

    return eofs, projections, variance_explained, mode_count



from .coast import Coast
from .coast import setup_dask_client
from .mask_maker import MaskMaker
from .gridded import Gridded
from .transect import Transect, TransectF, TransectT
from .contour import Contour, ContourF, ContourT
from .eof import compute_eofs, compute_hilbert_eofs
from .internal_tide import InternalTide
from .climatology import Climatology
from . import logging_util
from . import general_utils
from . import plot_util
from . import crps_util
from .index import Indexed
from .profile import Profile
from .profile_analysis import ProfileAnalysis
from .track import Track
from .lagrangian import Lagrangian
from .oceanparcels import Oceanparcels
from .glider import Glider
from .argos import Argos
from .altimetry import Altimetry
from .timeseries import Timeseries
from .tidegauge import Tidegauge
from .tidegauge_analysis import TidegaugeAnalysis
from .config_parser import ConfigParser
from .xesmf_convert import xesmf_convert
from .process_data import Process_data

# Set default for logging level when coast is imported
import logging

logging_util.setup_logging(level=logging.CRITICAL)



from .gridded import Gridded
from .logging_util import info, warn, error
import dask.array as da
from dask import delayed
import xarray as xr
import numpy as np


class Process_data:  # TODO All abstract methods should be implemented
    """
    A Python class containing methods for processing time series.
    """

    def __init__(self):
        return  # TODO Super __init__ should be called at some point

    def seasonal_decomp(self, ts_chunk, **kwargs):
        """
        Note this is intended as an inner functions to be called by a wrapper function
        Process_data.seasonal_decomposition

        This function is itself a wrapper function for statsmodel.seasonal_decompose
        that accepts multiple timeseries as multiple columns (s_dim)
        as an xr.DataArray: ts_chunk = ts_chunk(t_dim,s_dim).
        Returns the trend, seasonal and residual componenets of the time series.
        Invalid points i.e. land points, should be specified as np.nans.

        If called directly, evaluation is eager, i.e. numpy arrays are returned
        for the trend, seasonal and residual components of the time series

        ts_chunk is the input array or chunk
        **kwargs are the kwargs for statsmodels.seasonal_decompose

        Calling with dask.delayed:
        dask.delayed(self.seasonal_decomp, nout=3)(ts_chunk, **kwargs)

        Need statsmodels package installed
        """
        from statsmodels.tsa.seasonal import seasonal_decompose

        # Get the indices of land points (points that are nan for all time steps)
        active_ind = np.where(np.nan_to_num(ts_chunk.data).any(axis=0))[0]
        # Remove them (statsmodel.seasonal_decompose will not accept them)
        ts_chunk_valid = ts_chunk[:, active_ind]
        # Call statsmodels.seasonal_decompose()
        decomp_chunk = seasonal_decompose(ts_chunk_valid, **kwargs)
        # Store time series components in seperate arrays and return
        trend_chunk, seasonal_chunk, residual_chunk = (
            np.full(ts_chunk.shape, np.nan),
            np.full(ts_chunk.shape, np.nan),
            np.full(ts_chunk.shape, np.nan),
        )
        trend_chunk[:, active_ind] = decomp_chunk.trend
        seasonal_chunk[:, active_ind] = decomp_chunk.seasonal
        residual_chunk[:, active_ind] = decomp_chunk.resid
        return trend_chunk, seasonal_chunk, residual_chunk

    def seasonal_decomposition(self, time_series: xr.DataArray, num_chunks=1, **kwargs):
        """
        Dask delayed wrapper for statsmodel.seasonal_decompose that accepts multiple timeseries
        distributed across multiple dimensitons as an xr.DataArray:
        time_series = time_series(t_dim,z_dim,x_dim,y_dim,...)

        Accepts land points (z,x,y) as time_series=np.nan for all time.

        num_chunks allows caller to specify the number of chunks in the delayed call, which enables
        parallelism. i.e. num_chunks = 4 splits the timeserieses into 4 groups that will be decomposed
        in parallel by the dask scheduler.

        **kwargs are the kwargs for statsmodels.seasonal_decompose function

        Returns coast.Gridded object containing an xarray.Dataset with the trend, seasonal, residual
        timeseries components as xarray.DataArrays.

        Need statsmodels package installed
        """
        # first dim needs to be time
        time_series = time_series.transpose("t_dim", ...)
        # stack all the other dims
        ts_stacked = time_series.stack(space=time_series.dims[1:])
        # calculate chunk size for stacked dimension and rechunk
        chunk_size = ts_stacked["space"].size // num_chunks + 1
        ts_stacked = ts_stacked.chunk({"t_dim": ts_stacked["t_dim"].size, "space": chunk_size})
        # convert to list of dask delayed objects (e.g. 4 chunks => 4 delayed objects in list)
        ts_delayed = ts_stacked.data.to_delayed().ravel()

        # delayed call to statsmodel.seasonal_decomposition (with some other processing) that returns
        # 3 lists:
        # 1) contains the dask delayed chunks for trend (e.g. 4 delayed objects in list if 4 chunks),
        # 2) contains the dask delayed chunks for seasonal,
        # 3) contains the dask delayed chunks for residual,
        trend_de, seasonal_de, resid_de = map(
            list, zip(*[delayed(self.seasonal_decomp, nout=3)(ts_chunk, **kwargs) for ts_chunk in ts_delayed])
        )

        # convert to lists of dask arrays to allow array operations
        trend_da, seasonal_da, resid_da = [], [], []
        for chunk_idx, (tr_chunk, se_chunk, re_chunk) in enumerate(zip(trend_de, seasonal_de, resid_de)):
            # When converting from delayed to array, you must know the shape of the
            # array. Here we know this from the chunk sizes of the (stacked) DataArray
            chunk_shape = (ts_stacked.chunks[0][0], ts_stacked.chunks[1][chunk_idx])
            trend_da.append(da.from_delayed(tr_chunk, shape=chunk_shape, dtype=float))
            seasonal_da.append(da.from_delayed(se_chunk, shape=chunk_shape, dtype=float))
            resid_da.append(da.from_delayed(re_chunk, shape=chunk_shape, dtype=float))

        # concatenate the array chunks together and reshape back to original shape
        trend = da.reshape(da.concatenate(trend_da, axis=1), time_series.shape)
        seasonal = da.reshape(da.concatenate(seasonal_da, axis=1), time_series.shape)
        residual = da.reshape(da.concatenate(resid_da, axis=1), time_series.shape)
        # create  gridded object for return. New DataArrays created from the original
        # timeseries DataArray so it will have the smae coordinates and shape. Rechunk it
        # to have same structure as the newly created dask arrays and then assign to gridded
        gd = Gridded()
        gd.dataset["trend"] = xr.full_like(time_series.chunk(trend.chunks), np.nan)
        gd.dataset["trend"][:] = trend
        gd.dataset["seasonal"] = xr.full_like(time_series.chunk(seasonal.chunks), np.nan)
        gd.dataset["seasonal"][:] = seasonal
        gd.dataset["residual"] = xr.full_like(time_series.chunk(residual.chunks), np.nan)
        gd.dataset["residual"][:] = residual
        return gd
